#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2007  B. Malengier
# Copyright (C) 2008  Brian G. Matherly
# Copyright (C) 2011       Tim G L Lyons
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
This module provides the functions to build the quick report context menu's
"""

# ------------------------------------------------------------------------
#
# python modules
#
# ------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext
from io import StringIO

# ------------------------------------------------------------------------
#
# Set up logging
#
# ------------------------------------------------------------------------
import logging
from collections import abc

log = logging.getLogger(".quickreports")

# -------------------------------------------------------------------------
#
# GNOME modules
#
# -------------------------------------------------------------------------
from gi.repository import Gtk

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from ...pluginmanager import GuiPluginManager
from ...uimanager import valid_action_name
from gramps.gen.plug import (
    CATEGORY_QR_PERSON,
    CATEGORY_QR_FAMILY,
    CATEGORY_QR_MEDIA,
    CATEGORY_QR_EVENT,
    CATEGORY_QR_SOURCE,
    CATEGORY_QR_MISC,
    CATEGORY_QR_PLACE,
    CATEGORY_QR_REPOSITORY,
    CATEGORY_QR_NOTE,
    CATEGORY_QR_CITATION,
    CATEGORY_QR_SOURCE_OR_CITATION,
)
from ._textbufdoc import TextBufDoc
from gramps.gen.simple import make_basic_stylesheet

MENUITEM = (
    "<item>\n"
    '<attribute name="action">{prefix}.{action}</attribute>\n'
    '<attribute name="label">'
    "{label}</attribute>\n"
    "</item>\n"
)


def flatten(L):
    """
    Flattens a possibly nested list. Removes None results, too.
    """
    retval = []
    if isinstance(L, (list, tuple)):
        for item in L:
            fitem = flatten(item)
            if fitem is not None:
                retval.extend(fitem)
    elif L is not None:
        retval.append(L)
    return retval


def create_web_connect_menu(dbstate, uistate, nav_group, handle, prefix):
    """
    This functions querries the registered web connects.  It collects
    the connects of the requested category, which must be one of
    nav_group.

    It constructs the ui string of the menu, and it's actions. The
    action callback function is constructed, using the dbstate and the
    handle as input method.  A tuple is returned, containing the ui
    string of the menu, and its associated actions.
    """
    top = (
        "<placeholder id='WebConnect'><submenu>\n"
        '<attribute name="label" translatable="yes">'
        "Web Connection</attribute>\n"
    )
    ofile = StringIO()
    ofile.write(top)
    # select the web connects to show
    pmgr = GuiPluginManager.get_instance()
    plugins = pmgr.process_plugin_data("WebConnect")
    try:
        connections = [
            plug(nav_group) if isinstance(plug, abc.Callable) else plug
            for plug in plugins
        ]
    except BaseException:
        import traceback

        traceback.print_exc()
        connections = []
    connections = flatten(connections)
    connections.sort(key=lambda plug: plug.name)
    actions = []
    for indx, connect in enumerate(connections):
        # action would be better with "connect.key", but it seems to be
        # non-ASCII sometimes.  So we use an action number instead.
        action = "web-con-%d" % indx
        ofile.write(MENUITEM.format(prefix=prefix, action=action, label=connect.name))
        callback = connect(dbstate, uistate, nav_group, handle)
        actions.append((action, make_web_connect_callback(callback)))
    ofile.write("</submenu></placeholder>\n")
    return (ofile.getvalue(), actions)


def make_web_connect_callback(func):
    return lambda x, y: func(x)


def create_quickreport_menu(category, dbstate, uistate, handle, prefix, track=[]):
    """This functions querries the registered quick reports with
        quick_report_list of _PluginMgr.py
    It collects the reports of the requested category, which must be one of
                    CATEGORY_QR_PERSON, CATEGORY_QR_FAMILY,
                    CATEGORY_QR_EVENT, CATEGORY_QR_SOURCE, CATEGORY_QR_MEDIA,
                    CATEGORY_QR_PLACE, CATEGORY_QR_REPOSITORY,
                    CATEGORY_QR_CITATION, CATEGORY_QR_SOURCE_OR_CITATION
    It constructs the ui string of the quick report menu, and it's actions
    The action callback function is constructed, using the dbstate and the
        handle as input method.
    A tuple is returned, containing the ui string of the quick report menu,
    and its associated actions
    """
    top = (
        "<submenu>\n"
        '<attribute name="label" translatable="yes">'
        "Quick View</attribute>\n"
    )
    actions = []
    ofile = StringIO()
    ofile.write(top)

    # select the reports to show
    showlst = []
    pmgr = GuiPluginManager.get_instance()
    for pdata in pmgr.get_reg_quick_reports():
        if pdata.supported and pdata.category == category:
            showlst.append(pdata)

    showlst.sort(key=lambda x: x.name)
    for pdata in showlst:
        new_key = valid_action_name("qr-%s" % pdata.id)
        ofile.write(MENUITEM.format(prefix=prefix, action=new_key, label=pdata.name))
        actions.append(
            (
                new_key,
                make_quick_report_callback(
                    pdata, category, dbstate, uistate, handle, track=track
                ),
            )
        )
    ofile.write("</submenu>\n")

    return (ofile.getvalue(), actions)


def make_quick_report_callback(pdata, category, dbstate, uistate, handle, track=[]):
    return lambda x, y: run_report(
        dbstate, uistate, category, handle, pdata, track=track
    )


def get_quick_report_list(qv_category=None):
    """
    Returns a list of PluginData of quick views of category qv_category
    CATEGORY_QR_PERSON, CATEGORY_QR_FAMILY, CATEGORY_QR_EVENT,
    CATEGORY_QR_SOURCE, CATEGORY_QR_MISC, CATEGORY_QR_PLACE,
    CATEGORY_QR_REPOSITORY, CATEGORY_QR_MEDIA,
    CATEGORY_QR_CITATION, CATEGORY_QR_SOURCE_OR_CITATION or None for all
    """
    names = []
    pmgr = GuiPluginManager.get_instance()
    for pdata in pmgr.get_reg_quick_reports():
        if qv_category == pdata.category or qv_category is None:
            names.append(pdata)  # (see below for item struct)
    return names


def run_quick_report_by_name(
    dbstate, uistate, report_name, handle, container=None, track=[], **kwargs
):
    """
    Run a QuickView by name.
    **kwargs provides a way of passing special quick views additional
    arguments.
    """
    report = None
    pmgr = GuiPluginManager.get_instance()
    for pdata in pmgr.get_reg_quick_reports():
        if pdata.id == report_name:
            report = pdata
            break
    if report:
        return run_report(
            dbstate,
            uistate,
            report.category,
            handle,
            report,
            container=container,
            track=track,
            **kwargs,
        )
    else:
        raise AttributeError("No such quick report '%s'" % report_name)


def run_quick_report_by_name_direct(report_name, database, document, handle):
    """
    Useful for running one quick report from another
    """
    report = None
    pmgr = GuiPluginManager.get_instance()
    for pdata in pmgr.get_reg_quick_reports():
        if pdata.id == report_name:
            report = pdata
            break
    if report:
        # FIXME: allow auto lookup of obj like below?
        d = TextBufDoc(make_basic_stylesheet(), None)
        d.dbstate = document.dbstate
        d.uistate = document.uistate
        d.open("")
        mod = pmgr.load_plugin(report)
        if mod:
            reportfunc = getattr(mod, report.runfunc)
            retval = reportfunc(database, d, handle)
            d.close()
            return retval
        else:
            raise ImportError(
                "Quick report id = '%s' could not be loaded" % report_name
            )
    else:
        raise AttributeError("No such quick report id = '%s'" % report_name)


def run_report(
    dbstate, uistate, category, handle, pdata, container=None, track=[], **kwargs
):
    """
    Run a Quick Report.
    Optionally container can be passed, rather than putting the report
    in a new window.
    **kwargs are only used for special quick views that allow additional
    arguments, and that are run by run_quick_report_by_name().
    """
    pmgr = GuiPluginManager.get_instance()
    mod = pmgr.load_plugin(pdata)
    if not mod:
        print("QuickView Error: plugin does not load")
        return
    func = getattr(mod, pdata.runfunc)
    if handle:
        d = TextBufDoc(make_basic_stylesheet(), None, track=track)
        d.dbstate = dbstate
        d.uistate = uistate
        if isinstance(handle, str):  # a handle
            if category == CATEGORY_QR_PERSON:
                obj = dbstate.db.get_person_from_handle(handle)
            elif category == CATEGORY_QR_FAMILY:
                obj = dbstate.db.get_family_from_handle(handle)
            elif category == CATEGORY_QR_EVENT:
                obj = dbstate.db.get_event_from_handle(handle)
            elif category == CATEGORY_QR_SOURCE:
                obj = dbstate.db.get_source_from_handle(handle)
            elif category == CATEGORY_QR_CITATION:
                obj = dbstate.db.get_citation_from_handle(handle)
            elif category == CATEGORY_QR_SOURCE_OR_CITATION:
                if dbstate.db.has_source_handle(handle):
                    obj = dbstate.db.get_source_from_handle(handle)
                elif dbstate.db.has_citation_handle(handle):
                    obj = dbstate.db.get_citation_from_handle(handle)
                else:
                    raise ValueError("selection must be either source or citation")
            elif category == CATEGORY_QR_PLACE:
                obj = dbstate.db.get_place_from_handle(handle)
            elif category == CATEGORY_QR_MEDIA:
                obj = dbstate.db.get_media_from_handle(handle)
            elif category == CATEGORY_QR_REPOSITORY:
                obj = dbstate.db.get_repository_from_handle(handle)
            elif category == CATEGORY_QR_NOTE:
                obj = dbstate.db.get_note_from_handle(handle)
            elif category == CATEGORY_QR_MISC:
                obj = handle
            else:
                obj = None
        else:  # allow caller to send object directly
            obj = handle
        if obj:
            if container:
                result = d.open("", container=container)
                func(dbstate.db, d, obj, **kwargs)
                return result
            else:
                d.open("")
                retval = func(dbstate.db, d, obj, **kwargs)
                d.close()
                return retval
        else:
            print("QuickView Error: failed to run report: no obj")
    else:
        print("QuickView Error: handle is not set")



# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2001-2007  Donald N. Allingham
# Copyright (C) 2008       Gary Burton
# Copyright (C) 2011       Tim G L Lyons
# Copyright (C) 2017       Alois Poettker
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Provide the event view.
"""

# -------------------------------------------------------------------------
#
# Standard python modules
#
# -------------------------------------------------------------------------
import logging

_LOG = logging.getLogger(".plugins.eventview")

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.sgettext

from gramps.gui.dialog import ErrorDialog
from gramps.gen.errors import WindowActiveError
from gramps.gen.lib import Event

from gramps.gui.ddtargets import DdTargets
from gramps.gui.editors import EditEvent
from gramps.gui.filters.sidebar import EventSidebarFilter
from gramps.gui.merge import MergeEvent
from gramps.gen.plug import CATEGORY_QR_EVENT
from gramps.gui.views.bookmarks import EventBookmarks
from gramps.gui.views.listview import ListView, TEXT, MARKUP, ICON
from gramps.gui.views.treemodels import EventModel


# -------------------------------------------------------------------------
#
# EventView
#
# -------------------------------------------------------------------------
class EventView(ListView):
    """
    EventView class, derived from the ListView
    """

    # columns in the model used in view
    COL_DESCR = 0
    COL_ID = 1
    COL_TYPE = 2
    COL_DATE = 3
    COL_PLACE = 4
    COL_PRIV = 5
    COL_TAGS = 6
    COL_CHAN = 7
    COL_PARTIC = 8
    # column definitions
    COLUMNS = [
        (_("Description"), TEXT, None),
        (_("ID"), TEXT, None),
        (_("Type"), TEXT, None),
        (_("Date"), MARKUP, None),
        (_("Place"), TEXT, None),
        (_("Private"), ICON, "gramps-lock"),
        (_("Tags"), TEXT, None),
        (_("Last Changed"), TEXT, None),
        (_("Main Participants"), TEXT, None),
    ]
    # default setting with visible columns, order of the col, and their size
    CONFIGSETTINGS = (
        (
            "columns.visible",
            [COL_TYPE, COL_PARTIC, COL_DATE, COL_PLACE, COL_DESCR, COL_ID],
        ),
        (
            "columns.rank",
            [
                COL_TYPE,
                COL_PARTIC,
                COL_DATE,
                COL_PLACE,
                COL_DESCR,
                COL_ID,
                COL_PRIV,
                COL_TAGS,
                COL_CHAN,
            ],
        ),
        ("columns.size", [100, 230, 150, 200, 100, 75, 40, 100, 100]),
    )
    ADD_MSG = _("Add a new event")
    EDIT_MSG = _("Edit the selected event")
    DEL_MSG = _("Delete the selected event")
    MERGE_MSG = _("Merge the selected events")
    FILTER_TYPE = "Event"
    QR_CATEGORY = CATEGORY_QR_EVENT

    def __init__(self, pdata, dbstate, uistate, nav_group=0):
        """
        Create the Event View
        """
        signal_map = {
            "event-add": self.row_add,
            "event-update": self.row_update,
            "event-delete": self.row_delete,
            "event-rebuild": self.object_build,
            "person-update": self.person_update,
            "person-add": self.person_update,
            "person-delete": self.object_build,  # TODO slow way to do this
            "family-update": self.family_update,
            "family-add": self.family_update,
            "family-delete": self.object_build,  # TODO slow way to do this
            "place-update": self.related_update,
        }

        ListView.__init__(
            self,
            _("Events"),
            pdata,
            dbstate,
            uistate,
            EventModel,
            signal_map,
            EventBookmarks,
            nav_group,
            multiple=True,
            filter_class=EventSidebarFilter,
        )

        uistate.connect("nameformat-changed", self.build_tree)
        uistate.connect("placeformat-changed", self.build_tree)

        self.additional_uis.append(self.additional_ui)

    def person_update(self, hndl_list):
        """Deal with person updates thay may effect the Main Participants
        column.  These cannot use the more generic mechanism because Person
        objects use EventRef to point to events, rather than Events pointing
        to persons. Example: A person's name change or add event to person"""
        update_list = []
        for hndl in hndl_list:
            person = self.dbstate.db.get_person_from_handle(hndl)
            for eventref in person.get_event_ref_list():
                update_list.append(eventref.ref)
        self.row_update(update_list)

    def family_update(self, hndl_list):
        """Deal with family updates thay may effect the Main Participants
        column.  These cannot use the more generic mechanism because Family
        objects use EventRef to point to events, rather than Events pointing
        to Families. Example: Change/add/removal of parent, or add family to
        event"""
        update_list = []
        for hndl in hndl_list:
            family = self.dbstate.db.get_family_from_handle(hndl)
            for eventref in family.get_event_ref_list():
                update_list.append(eventref.ref)
        self.row_update(update_list)

    def navigation_type(self):
        return "Event"

    def drag_info(self):
        """
        Indicate that the drag type is an EVENT
        """
        return DdTargets.EVENT

    def get_stock(self):
        """
        Use the gramps-event stock icon
        """
        return "gramps-event"

    additional_ui = [  # Defines the UI string for UIManager
        """
      <placeholder id="LocalExport">
        <item>
          <attribute name="action">win.ExportTab</attribute>
          <attribute name="label" translatable="yes">Export View...</attribute>
        </item>
      </placeholder>
""",
        """
      <section id="AddEditBook">
        <item>
          <attribute name="action">win.AddBook</attribute>
          <attribute name="label" translatable="yes">_Add Bookmark</attribute>
        </item>
        <item>
          <attribute name="action">win.EditBook</attribute>
          <attribute name="label" translatable="no">%s...</attribute>
        </item>
      </section>
"""
        % _("Organize Bookmarks"),
        """
      <placeholder id="CommonGo">
      <section>
        <item>
          <attribute name="action">win.Back</attribute>
          <attribute name="label" translatable="yes">_Back</attribute>
        </item>
        <item>
          <attribute name="action">win.Forward</attribute>
          <attribute name="label" translatable="yes">_Forward</attribute>
        </item>
      </section>
      </placeholder>
""",
        """
      <section id='CommonEdit' groups='RW'>
        <item>
          <attribute name="action">win.Add</attribute>
          <attribute name="label" translatable="yes">_Add...</attribute>
        </item>
        <item>
          <attribute name="action">win.Edit</attribute>
          <attribute name="label">%s</attribute>
        </item>
        <item>
          <attribute name="action">win.Remove</attribute>
          <attribute name="label" translatable="yes">_Delete</attribute>
        </item>
        <item>
          <attribute name="action">win.Merge</attribute>
          <attribute name="label" translatable="yes">_Merge...</attribute>
        </item>
      </section>
"""
        % _("_Edit...", "action"),  # to use sgettext()
        """
        <placeholder id='otheredit'>
        <item>
          <attribute name="action">win.FilterEdit</attribute>
          <attribute name="label" translatable="yes">"""
        """Event Filter Editor</attribute>
        </item>
        </placeholder>
""",  # Following are the Toolbar items
        """
    <placeholder id='CommonNavigation'>
    <child groups='RO'>
      <object class="GtkToolButton">
        <property name="icon-name">go-previous</property>
        <property name="action-name">win.Back</property>
        <property name="tooltip_text" translatable="yes">"""
        """Go to the previous object in the history</property>
        <property name="label" translatable="yes">_Back</property>
        <property name="use-underline">True</property>
      </object>
      <packing>
        <property name="homogeneous">False</property>
      </packing>
    </child>
    <child groups='RO'>
      <object class="GtkToolButton">
        <property name="icon-name">go-next</property>
        <property name="action-name">win.Forward</property>
        <property name="tooltip_text" translatable="yes">"""
        """Go to the next object in the history</property>
        <property name="label" translatable="yes">_Forward</property>
        <property name="use-underline">True</property>
      </object>
      <packing>
        <property name="homogeneous">False</property>
      </packing>
    </child>
    </placeholder>
""",
        """
    <placeholder id='BarCommonEdit'>
    <child groups='RW'>
      <object class="GtkToolButton">
        <property name="icon-name">list-add</property>
        <property name="action-name">win.Add</property>
        <property name="tooltip_text">%s</property>
        <property name="label" translatable="yes">_Add...</property>
        <property name="use-underline">True</property>
      </object>
      <packing>
        <property name="homogeneous">False</property>
      </packing>
    </child>
    <child groups='RW'>
      <object class="GtkToolButton">
        <property name="icon-name">gtk-edit</property>
        <property name="action-name">win.Edit</property>
        <property name="tooltip_text">%s</property>
        <property name="label" translatable="yes">Edit...</property>
        <property name="use-underline">True</property>
      </object>
      <packing>
        <property name="homogeneous">False</property>
      </packing>
    </child>
    <child groups='RW'>
      <object class="GtkToolButton">
        <property name="icon-name">list-remove</property>
        <property name="action-name">win.Remove</property>
        <property name="tooltip_text">%s</property>
        <property name="label" translatable="yes">_Delete</property>
        <property name="use-underline">True</property>
      </object>
      <packing>
        <property name="homogeneous">False</property>
      </packing>
    </child>
    <child groups='RW'>
      <object class="GtkToolButton">
        <property name="icon-name">gramps-merge</property>
        <property name="action-name">win.Merge</property>
        <property name="tooltip_text">%s</property>
        <property name="label" translatable="yes">_Merge...</property>
        <property name="use-underline">True</property>
      </object>
      <packing>
        <property name="homogeneous">False</property>
      </packing>
    </child>
   </placeholder>
"""
        % (ADD_MSG, EDIT_MSG, DEL_MSG, MERGE_MSG),
        """
    <menu id="Popup">
      <section>
        <item>
          <attribute name="action">win.Back</attribute>
          <attribute name="label" translatable="yes">_Back</attribute>
        </item>
        <item>
          <attribute name="action">win.Forward</attribute>
          <attribute name="label" translatable="yes">Forward</attribute>
        </item>
      </section>
      <section id="PopUpTree">
      </section>
      <section>
        <item>
          <attribute name="action">win.Add</attribute>
          <attribute name="label" translatable="yes">_Add...</attribute>
        </item>
        <item>
          <attribute name="action">win.Edit</attribute>
          <attribute name="label">%s</attribute>
        </item>
        <item>
          <attribute name="action">win.Remove</attribute>
          <attribute name="label" translatable="yes">_Delete</attribute>
        </item>
        <item>
          <attribute name="action">win.Merge</attribute>
          <attribute name="label" translatable="yes">_Merge...</attribute>
        </item>
      </section>
      <section>
        <placeholder id='QuickReport'>
        </placeholder>
        <placeholder id='WebConnect'>
        </placeholder>
      </section>
    </menu>
"""
        % _("_Edit...", "action"),  # to use sgettext()
    ]

    def get_handle_from_gramps_id(self, gid):
        obj = self.dbstate.db.get_event_from_gramps_id(gid)
        if obj:
            return obj.get_handle()
        else:
            return None

    def add(self, *obj):
        try:
            EditEvent(self.dbstate, self.uistate, [], Event())
        except WindowActiveError:
            pass

    def remove(self, *obj):
        """
        Method called when deleting event(s) from the event view.
        """
        handles = self.selected_handles()
        ht_list = [("Event", hndl) for hndl in handles]
        self.remove_selected_objects(ht_list)

    def _message1_format(self, event):
        """
        Header format for remove dialogs.
        """
        return _("Delete {type} [{gid}]?").format(
            type=str(event.type), gid=event.gramps_id
        )

    def edit(self, *obj):
        for handle in self.selected_handles():
            event = self.dbstate.db.get_event_from_handle(handle)
            try:
                EditEvent(self.dbstate, self.uistate, [], event)
            except WindowActiveError:
                pass

    def merge(self, *obj):
        """
        Merge the selected events.
        """
        mlist = self.selected_handles()

        if len(mlist) != 2:
            msg = _("Cannot merge event objects.")
            msg2 = _(
                "Exactly two events must be selected to perform a merge. "
                "A second object can be selected by holding down the "
                "control key while clicking on the desired event."
            )
            ErrorDialog(msg, msg2, parent=self.uistate.window)
        else:
            MergeEvent(self.dbstate, self.uistate, [], mlist[0], mlist[1])

    def tag_updated(self, handle_list):
        """
        Update tagged rows when a tag color changes.
        """
        all_links = set([])
        for tag_handle in handle_list:
            links = set(
                [
                    link[1]
                    for link in self.dbstate.db.find_backlink_handles(
                        tag_handle, include_classes="Event"
                    )
                ]
            )
            all_links = all_links.union(links)
        self.row_update(list(all_links))

    def add_tag(self, transaction, event_handle, tag_handle):
        """
        Add the given tag to the given event.
        """
        event = self.dbstate.db.get_event_from_handle(event_handle)
        event.add_tag(tag_handle)
        self.dbstate.db.commit_event(event, transaction)

    def remove_tag(self, transaction, event_handle, tag_handle):
        """
        Remove the given tag from the given event.
        """
        event = self.dbstate.db.get_event_from_handle(event_handle)
        event.remove_tag(tag_handle)
        self.dbstate.db.commit_event(event, transaction)

    def get_default_gramplets(self):
        """
        Define the default gramplets for the sidebar and bottombar.
        """
        return (
            ("Event Filter",),
            (
                "Event Gallery",
                "Event Citations",
                "Event Notes",
                "Event Attributes",
                "Event Backlinks",
            ),
        )

    def get_config_name(self):
        return __name__



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2001-2006 Donald N. Allingham
# Copyright (C)      2010 Rob G. Healey <robhealey1@gmail.com>
# Copyright (C)      2010 Jakim Friant
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"Report Generation Framework"

# -------------------------------------------------------------------------
#
# standard python modules
#
# -------------------------------------------------------------------------
from ...const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext
import os

# -------------------------------------------------------------------------
#
# Constants
#
# -------------------------------------------------------------------------

# Report categories
from .. import (
    CATEGORY_TEXT,
    CATEGORY_DRAW,
    CATEGORY_CODE,
    CATEGORY_WEB,
    CATEGORY_BOOK,
    CATEGORY_GRAPHVIZ,
    CATEGORY_TREE,
)

standalone_categories = {
    CATEGORY_TEXT: ("RepText", _("Text Reports")),
    CATEGORY_DRAW: ("RepGraph", _("Graphical Reports")),
    CATEGORY_CODE: ("RepCode", _("Code Generators")),
    CATEGORY_WEB: ("RepWeb", _("Web Pages")),
    CATEGORY_BOOK: ("RepBook", _("Books")),
    CATEGORY_GRAPHVIZ: ("Graphs", _("Graphs")),
    CATEGORY_TREE: ("Trees", _("Trees")),
}
book_categories = {
    CATEGORY_TEXT: _("Text"),
    CATEGORY_DRAW: _("Graphics"),
}



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2017   Alois Poettker
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"Import from Pro-Gen"

# -------------------------------------------------------------------------
#
# standard python modules
#
# -------------------------------------------------------------------------
import os, time
import gi

gi.require_version("Gtk", "3.0")
from gi.repository import Gtk, Gdk

# ------------------------------------------------------------------------
#
# Gramps modules
#
# ------------------------------------------------------------------------
from gramps.gen.const import URL_MANUAL_PAGE
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext

from gramps.gen.config import config

from gramps.gen.lib import Citation
from gramps.gen.lib.date import Today
from gramps.gen.lib.datebase import DateBase
from gramps.gen.utils.libformatting import ImportInfo

from gramps.gui.dialog import InfoDialog
from gramps.gui.display import display_help
from gramps.gui.glade import Glade
from gramps.gui.managedwindow import ManagedWindow
from gramps.gui.widgets import (
    MonitoredCheckbox,
    MonitoredDate,
    MonitoredEntry,
    MonitoredMenu,
    PrivacyButton,
)

from gramps.plugins.lib import libprogen

# -------------------------------------------------------------------------
#
# Constants
#
# -------------------------------------------------------------------------
WIKI_HELP_PAGE = "%s_-_Importer" % URL_MANUAL_PAGE
WIKI_HELP_SEC = _("Import_from_another_genealogy_program", "manual")


# -------------------------------------------------------------------------
#
# importData
#
# -------------------------------------------------------------------------
class ImportPrivacy(object):
    """Class for internal values of privacy objects"""

    def __init__(self, value):
        """"""
        self.priv_value = value

    # sets / gets Privacy value
    def set_privacy(self, value):
        """connects to 'set' method of PrivacyCheckbox"""
        self.priv_value = value

    def get_privacy(self):
        """connects to 'get' method of PrivacyCheckbox"""
        return self.priv_value


class ImportValue(object):
    """Class for internal values of objects"""

    def __init__(self, value):
        """"""
        self.value = value

    # sets / gets Button value
    def set_value(self, value):
        """connects to 'set' method of MonitoredCheckbox"""
        self.value = value

    def get_value(self):
        """connects to 'get' method of MonitoredCheckbox"""
        return self.value


class ImportEntry(object):
    """Class for internal entrys of objects"""

    def __init__(self, entry):
        """"""
        self.entry = entry

    # sets / gets Entry text
    def set_entry(self, entry):
        """connects to 'set' method of MonitoredEntry"""
        self.entry = entry

    def get_entry(self):
        """connects to 'get' method of MonitoredEntry"""
        return self.entry


class ImportSourceCitation(object):
    """Class for internal values for source / citation objects"""

    def __init__(self, dflt_btn, dflt_text, text, fname, date):
        """"""
        self.source_btn = ImportValue(dflt_btn)
        self.source_priv = ImportPrivacy(False)
        self.source_title = ImportEntry(_("Import from Pro-Gen (%s)") % fname)
        self.source_attr = ImportEntry("%s (%s) %s" % (text, fname, date))
        self.citation_btn = ImportValue(dflt_btn)
        self.citation_priv = ImportPrivacy(False)
        self.citation_conf = ImportValue(Citation.CONF_HIGH)
        self.citation_page = ImportEntry("%s" % dflt_text)
        self.citation_attr = ImportEntry("%s (%s) %s" % (text, fname, date))


class ImportTagTextDefault(object):
    """Class for internal values for default objects"""

    def __init__(self, text, fname):
        """"""
        self.text = ImportEntry(text)
        self.fname = ImportEntry(fname)
        self.date = Today()
        self.date.text = "{:04d}-{:02d}-{:02d}".format(
            self.date.dateval[2], self.date.dateval[1], self.date.dateval[0]
        )

    # sets / gets Default entries
    def set_dfltdate(self, date):
        """connects to 'set' method of MonitoredDate"""
        self.date = date

    def get_dfltdate(self):
        """connects to 'get' method of MonitoredDate"""
        return self.date.text


class ImportTagText(object):
    """Class for internal values for tag objects"""

    def __init__(self, default):
        """"""
        self.tag_dflt = default
        self.tag_obj = ImportValue(True)
        self.tag_fname = ImportValue(True)
        self.tag_date = ImportValue(True)
        tag_text = "%s (%s) %s" % (
            default.text.get_entry(),
            default.fname.get_entry(),
            default.get_dfltdate(),
        )
        self.tag_text = ImportEntry(tag_text)

    def get_dflttext(self):
        """set Tag Text to default values"""
        tag_text = self.tag_dflt.text.get_entry()
        if self.tag_fname.get_value():
            tag_text += " (%s)" % self.tag_dflt.fname.get_entry()
        if self.tag_date.get_value():
            tag_text += " %s" % self.tag_dflt.get_dfltdate()
        self.tag_text.set_entry(tag_text)

        return tag_text


def _importData(database, filename, user):
    """
    Imports the files corresponding to the specified Database, Filename & User.
    """
    data = ProgenOptions(database, filename, user)
    if data.fail:  # 'Cancel' button pressed
        return

    data = libprogen.ProgenParser(database, filename, user, data.option)
    try:
        info = libprogen.ProgenParser.parse_progen_file(data)
    except libprogen.ProgenError as msg:
        user.notify_error(_("Pro-Gen data error"), str(msg))
        return
    except IOError as msg:
        user.notify_error(_("%s could not be opened") % filename, str(msg))
        return

    if info:  # successful import
        # display qualified/standard statistic window
        if user.uistate:
            InfoDialog(_("Import Statistics"), info.info_text(), parent=user.parent)
        else:
            return ImportInfo({_("Results"): _("done")})


class ProgenOptions(ManagedWindow):
    """Class for Pro-Gen files import optons"""

    def __init__(self, database, filename, user):
        """Useful options for Pro-Gen import"""
        self.dbase = database
        self.fname = filename
        self.user = user
        self.uistate = user.uistate
        self.option = {}

        # default: Pro-Gen import failed
        self.fail = True

        # initial values
        text = "Pro-Gen Import"
        fname = os.path.basename(filename).split("\\")[-1]
        date = time.strftime("%Y-%m-%d", time.localtime())

        # add import source title/confidence
        #            citation page/confidence/privacy/attribute
        self.import_methods = {}
        dflt = config.get("preferences.default-source")
        if config.get("preferences.tag-on-import"):
            dflt_text = config.get("preferences.tag-on-import-format")
        else:
            dflt_text = "%s %s" % (text, date)
        self.imp_values = ImportSourceCitation(dflt, dflt_text, text, fname, date)

        # add default text / filename / current date
        self.default_methods = {}
        self.default_values = ImportTagTextDefault(text, fname)

        # add tagobjects text / filename / date
        self.tagobj_status = True
        self.tagobj_values, self.tagobj_methods = {}, {}
        self.tagtext_methods = {}
        self.tagfname_status, self.tagfname_methods = True, {}
        self.tagdate_status, self.tagdate_methods = True, {}
        for obj in libprogen.TAGOBJECTS:
            self.tagobj_values[obj] = ImportTagText(self.default_values)

        # add primary object values
        self.primobj_values, self.primobj_methods = {}, {}
        for obj in libprogen.PRIMOBJECTS:
            self.primobj_values[obj] = ImportValue(True)

        # add options values
        self.option_values, self.option_methods = {}, {}
        for obj in libprogen.OPTOBJECTS:
            if obj in ["person-ident", "family-ident", "surname-female", "death-cause"]:
                self.option_values[obj] = ImportValue(True)
            else:
                self.option_values[obj] = ImportValue(False)

        # prepare option dictionary
        self._collect()

        # display window if GUI active
        if self.uistate:
            ManagedWindow.__init__(self, self.uistate, [], self.__class__, modal=True)
            self._display()

    def __on_source_button_toggled(self, widget):
        """compute the source button and toggle the 'Sensitive' attribute"""
        obj_source_state = widget.get_active()
        for obj in ["priv", "title", "attr"]:
            imp_obj = self.glade.get_object("imp_source_%s" % obj)
            imp_obj.set_sensitive(obj_source_state)

        # Check if Source enabled and syncronizing Citation
        self.glade.get_object("imp_citation_btn").set_active(obj_source_state)
        self.glade.get_object("imp_citation_btn").set_sensitive(obj_source_state)

    def __on_citation_button_toggled(self, widget):
        """compute the citation button and toggle the 'Sensitive' attribute"""
        # Check if Source enabled and syncronizing Citation
        obj_source_state = self.glade.get_object("imp_source_btn").get_active()

        obj_citation_state = widget.get_active() and obj_source_state
        for obj in ["priv", "conf", "page", "attr"]:
            imp_obj = self.glade.get_object("imp_citation_%s" % obj)
            imp_obj.set_sensitive(obj_citation_state)

    def __on_import_entry_keyrelease(self, widget, event, data=None):
        """activated on all return's of an entry"""
        if event.keyval in [Gdk.KEY_Return]:
            obj_name = Gtk.Buildable.get_name(widget).split("_", 1)[1]
            if obj_name == "citation_page":
                obj_next = self.glade.get_object("imp_citation_attr")
                obj_next.grab_focus()

    def __on_object_button_clicked(self, widget=None):
        """compute all primary objects and toggle the 'Active' attribute"""
        self.tagobj_status = not self.tagobj_status

        for obj in libprogen.TAGOBJECTS:
            tag_obj = self.glade.get_object("tag_%s_obj" % obj)
            tag_obj.set_active(self.tagobj_status)

    def __on_object_button_toggled(self, widget):
        """compute the primary object and toggle the 'Sensitive' attribute"""
        obj_state = widget.get_active()
        obj_name = Gtk.Buildable.get_name(widget)
        obj_name = obj_name.split("_", 1)[1].split("_", 1)[0]

        for obj in ["text", "fname", "date"]:
            tag_obj = self.glade.get_object("tag_%s_%s" % (obj_name, obj))
            tag_obj.set_sensitive(obj_state)

    def __on_text_button_clicked(self, widget=None):
        """compute all primary objects and flush the 'text' field"""
        self.__on_tagtext_entry_resume()  # Resume tag text

    def __on_tagtext_entry_keyrelease(self, widget, event, data=None):
        """activated on all return's of an entry"""
        if event.keyval in [Gdk.KEY_Return]:
            obj_name = Gtk.Buildable.get_name(widget)
            obj_name = obj_name.split("_", 1)[1].split("_", 1)[0]
            obj_index = libprogen.TAGOBJECTS.index(obj_name)
            if obj_index < len(libprogen.TAGOBJECTS) - 1:
                obj_index = obj_index + 1
            obj_next = self.glade.get_object(
                "tag_%s_text" % libprogen.TAGOBJECTS[obj_index]
            )
            obj_next.grab_focus()

    def __on_tagtext_entry_resume(self):
        """resume new tagtext from old + file & date variables"""
        for obj in libprogen.TAGOBJECTS:
            tag_obj = self.glade.get_object("tag_%s_text" % obj)
            if not tag_obj.get_sensitive():
                continue

            obj_entry = self.tagobj_values[obj].get_dflttext()
            self.tagobj_values[obj].tag_text.set_entry(obj_entry)
            self.tagtext_methods[obj].update()

    def __on_fname_button_clicked(self, widget=None):
        """compute all primary objects and toggle the 'file' attribute"""
        self.tagfname_status = not self.tagfname_status

        for obj in libprogen.TAGOBJECTS:
            tag_obj = self.glade.get_object("tag_%s_fname" % obj)
            if not tag_obj.get_sensitive():
                continue

            self.tagfname_methods[obj].set_val(self.tagfname_status)
            tag_obj.set_active(self.tagfname_status)

        self.__on_tagtext_entry_resume()  # Resume tag text

    def __on_fname_button_toggled(self, widget):
        """compute the primary object and toggle the 'Sensitive' attribute"""
        self.__on_tagtext_entry_resume()

        # switch focus forward
        obj_name = Gtk.Buildable.get_name(widget)
        obj_name = obj_name.split("_", 1)[1].split("_", 1)[0]
        obj_index = libprogen.TAGOBJECTS.index(obj_name)
        if obj_index < len(libprogen.TAGOBJECTS) - 1:
            obj_index = obj_index + 1
        obj_next = self.glade.get_object(
            "tag_%s_fname" % libprogen.TAGOBJECTS[obj_index]
        )
        obj_next.grab_focus()

    def __on_date_button_clicked(self, widget=None):
        """compute all primary objects and toggle the 'date' attribute"""
        self.tagdate_status = not self.tagdate_status

        for obj in libprogen.TAGOBJECTS:
            tag_obj = self.glade.get_object("tag_%s_date" % obj)
            if not tag_obj.get_sensitive():
                continue

            self.tagdate_methods[obj].set_val(self.tagdate_status)
            tag_obj.set_active(self.tagdate_status)

        self.__on_tagtext_entry_resume()  # Resume tag text

    def __on_date_button_toggled(self, widget):
        """compute the primary object and toggle the 'Sensitive' attribute"""
        self.__on_tagtext_entry_resume()

        # switch focus forward
        obj_name = Gtk.Buildable.get_name(widget)
        obj_name = obj_name.split("_", 1)[1].split("_", 1)[0]
        obj_index = libprogen.TAGOBJECTS.index(obj_name)
        if obj_index < len(libprogen.TAGOBJECTS) - 1:
            obj_index = obj_index + 1
        obj_next = self.glade.get_object(
            "tag_%s_date" % libprogen.TAGOBJECTS[obj_index]
        )
        obj_next.grab_focus()

    def __on_primobj_button_toggled(self, widget):
        """compute all primary objects and toggle the 'primobj' attribute"""
        obj_name = Gtk.Buildable.get_name(widget)
        obj_name = obj_name.split("_", 1)[1].split("_", 1)[0]

        for i, obj_i in enumerate(libprogen.PRIMOBJECTS):
            if obj_name == libprogen.PRIMOBJECTS[i]:
                prim_obj = self.glade.get_object("prim_%s_btn" % obj_i)
                if not prim_obj.get_active():
                    # Prim. object deactivated:
                    # check all in list below and deactivate
                    for obj_ii in libprogen.PRIMOBJECTS[(i + 1) :]:
                        sec_obj = self.glade.get_object("prim_%s_btn" % obj_ii)
                        sec_obj.set_active(False)
                    break
                else:
                    # Prim. object activated: check all in list above
                    # if one prim. object is deactive:
                    # deactivate this prim. object too
                    for obj_ii in libprogen.PRIMOBJECTS[:i]:
                        sec_obj = self.glade.get_object("prim_%s_btn" % obj_ii)
                        if not sec_obj.get_active():
                            prim_obj.set_active(False)
                            break

    def __on_surname_button_toggled(self, widget):
        """compute surname objects and"""
        obj_name = Gtk.Buildable.get_name(widget)
        obj_name = obj_name.split("_", 1)[1].split("_", 1)[0]
        obj_status = widget.get_active()

        if (obj_name == "name-surmale") and (obj_status == True):
            sec_obj = self.glade.get_object("opt_surname-female_btn")
            sec_obj.set_active(False)
        if (obj_name == "name-surfemale") and (obj_status == True):
            sec_obj = self.glade.get_object("opt_surname-male_btn")
            sec_obj.set_active(False)

    def __on_x_button_clicked(self, widget=None):
        """cancel the import and close"""
        self.fail = True  # Pro-Gen import canceled

    def __on_ok_button_clicked(self, widget=None):
        """execute the import and close"""
        self._collect()
        self.close()

        self.fail = False  # Pro-Gen import proceed

    def __on_cancel_button_clicked(self, widget=None):
        """cancel the import and close"""
        self.close()

        self.fail = True  # Pro-Gen import canceled

    def __on_help_button_clicked(self, widget=None):
        """display the relevant portion of Gramps manual"""
        self.fail = True

        display_help(webpage=WIKI_HELP_PAGE, section=WIKI_HELP_SEC)

    def build_menu_names_(self, widget=None):
        """The menu name"""
        return (_("Main window"), _("Import Pro-Gen"))

    def _display(self):
        """organize Glade 'Import Pro-Gen' window"""

        # get the main window from glade
        self.glade = Glade("importprogen.glade")
        self.set_window(
            self.glade.toplevel, self.glade.get_object("title"), _("Import Pro-Gen")
        )

        # calculate all entries and update Glade window
        # Text for Source / Citation objects
        for obj in ("source_btn", "citation_btn"):
            widget = self.glade.get_object("imp_%s" % obj)
            set_import = eval("self.imp_values.%s.set_value" % obj)
            get_import = eval("self.imp_values.%s.get_value" % obj)
            self.import_methods[obj] = MonitoredCheckbox(
                widget, widget, set_import, get_import, self.dbase.readonly
            )
        for obj in ("source_title", "source_attr", "citation_page", "citation_attr"):
            widget = self.glade.get_object("imp_%s" % obj)
            set_import = eval("self.imp_values.%s.set_entry" % obj)
            get_import = eval("self.imp_values.%s.get_entry" % obj)
            self.import_methods[obj] = MonitoredEntry(
                widget, set_import, get_import, self.dbase.readonly
            )
        widget = self.glade.get_object("imp_citation_conf")
        self.import_methods["conf"] = MonitoredMenu(
            widget,
            self.imp_values.citation_conf.set_value,
            self.imp_values.citation_conf.get_value,
            [
                (_("Very Low"), Citation.CONF_VERY_LOW),
                (_("Low"), Citation.CONF_LOW),
                (_("Normal"), Citation.CONF_NORMAL),
                (_("High"), Citation.CONF_HIGH),
                (_("Very High"), Citation.CONF_VERY_HIGH),
            ],
            self.dbase.readonly,
        )
        widget = self.glade.get_object("imp_source_priv")
        get_import = eval("self.imp_values.source_priv")
        self.import_methods["source_priv"] = PrivacyButton(
            widget, get_import, self.dbase.readonly
        )
        widget = self.glade.get_object("imp_citation_priv")
        get_import = eval("self.imp_values.citation_priv")
        self.import_methods["citation_priv"] = PrivacyButton(
            widget, get_import, self.dbase.readonly
        )

        # Text (w. Defaults) for Tags
        for obj in ("text", "fname"):
            widget = self.glade.get_object("tag_default_%s" % obj)
            set_import = eval("self.default_values.%s.set_entry" % obj)
            get_import = eval("self.default_values.%s.get_entry" % obj)
            self.default_methods[obj] = MonitoredEntry(
                widget, set_import, get_import, self.dbase.readonly
            )
        date = Today()
        datebase = DateBase()
        datebase.set_date_object(date)
        self.default_methods["date"] = MonitoredDate(
            self.glade.get_object("tag_default_date"),
            self.glade.get_object("tag_default_date_btn"),
            datebase.get_date_object(),
            self.uistate,
            [],
            self.dbase.readonly,
        )

        for obj in libprogen.TAGOBJECTS:
            # populate object fields with values
            widget = self.glade.get_object("tag_%s_obj" % obj)
            self.tagobj_methods[obj] = MonitoredCheckbox(
                widget,
                widget,
                self.tagobj_values[obj].tag_obj.set_value,
                self.tagobj_values[obj].tag_obj.get_value,
            )
            widget = self.glade.get_object("tag_%s_text" % obj)
            self.tagtext_methods[obj] = MonitoredEntry(
                widget,
                self.tagobj_values[obj].tag_text.set_entry,
                self.tagobj_values[obj].tag_text.get_entry,
            )
            widget = self.glade.get_object("tag_%s_fname" % obj)
            self.tagfname_methods[obj] = MonitoredCheckbox(
                widget,
                widget,
                self.tagobj_values[obj].tag_fname.set_value,
                self.tagobj_values[obj].tag_fname.get_value,
            )
            widget = self.glade.get_object("tag_%s_date" % obj)
            self.tagdate_methods[obj] = MonitoredCheckbox(
                widget,
                widget,
                self.tagobj_values[obj].tag_date.set_value,
                self.tagobj_values[obj].tag_date.get_value,
            )

        # button's for primary objects
        for obj in libprogen.PRIMOBJECTS:
            # populate pirm. Object buttons with values
            widget = self.glade.get_object("prim_%s_btn" % obj)
            set_import = eval("self.primobj_values['%s'].set_value" % obj)
            get_import = eval("self.primobj_values['%s'].get_value" % obj)
            self.primobj_methods[obj] = MonitoredCheckbox(
                widget, widget, set_import, get_import, self.dbase.readonly
            )

        # button's for miscallaneous option's
        for obj in libprogen.OPTOBJECTS:
            # populate option buttons with values
            widget = self.glade.get_object("opt_%s_btn" % obj)
            set_import = eval("self.option_values['%s'].set_value" % obj)
            get_import = eval("self.option_values['%s'].get_value" % obj)
            self.option_methods[obj] = MonitoredCheckbox(
                widget, widget, set_import, get_import, self.dbase.readonly
            )

        # connect signals
        self.glade.connect_signals(
            {
                "on_source_button_toggled": self.__on_source_button_toggled,
                "on_citation_button_toggled": self.__on_citation_button_toggled,
                "on_import_entry_keyrelease": self.__on_import_entry_keyrelease,
                "on_tagtext_entry_keyrelease": self.__on_tagtext_entry_keyrelease,
                "on_object_button_clicked": self.__on_object_button_clicked,
                "on_object_button_toggled": self.__on_object_button_toggled,
                "on_text_button_clicked": self.__on_text_button_clicked,
                "on_fname_button_clicked": self.__on_fname_button_clicked,
                "on_fname_button_toggled": self.__on_fname_button_toggled,
                "on_date_button_clicked": self.__on_date_button_clicked,
                "on_date_button_toggled": self.__on_date_button_toggled,
                "on_primobj_button_toggled": self.__on_primobj_button_toggled,
                "on_surname_button_toggled": self.__on_surname_button_toggled,
                "on_x_button_clicked": self.__on_x_button_clicked,
                "on_help_button_clicked": self.__on_help_button_clicked,
                "on_cancel_button_clicked": self.__on_cancel_button_clicked,
                "on_ok_button_clicked": self.__on_ok_button_clicked,
            }
        )

        # state of two objects trigged form configuration
        widget = self.glade.get_object("imp_source_btn")
        self.__on_source_button_toggled(widget)
        widget = self.glade.get_object("imp_citation_btn")
        self.__on_citation_button_toggled(widget)
        widget = self.glade.get_object("import_ok")
        widget.grab_focus()

        # creates a modal window and display immediatly!
        self.show()
        self.glade.toplevel.run()

    def _collect(self):
        """collect all options"""

        self.option["imp_source"] = self.imp_values.source_btn.get_value()
        self.option["imp_source_priv"] = (
            self.imp_values.source_priv.get_privacy()
            if self.option["imp_source"]
            else False
        )
        if self.option["imp_source"]:
            self.option["imp_source_title"] = self.imp_values.source_title.get_entry()
            sourceattr = self.imp_values.source_attr.get_entry()
            # Expand if exists possible strftime directives
            if ("%Y" or "%m" or "%d" or "%H" or "%M" or "%S") in sourceattr:
                sourceattr = time.strftime(sourceattr)
            self.option["imp_source_attr"] = sourceattr
        else:
            self.option["imp_source_title"] = ""
            self.option["imp_source_attr"] = ""

        self.option["imp_citation"] = self.imp_values.citation_btn.get_value()
        self.option["imp_citation_conf"] = (
            self.imp_values.citation_conf.get_value()
            if self.option["imp_citation"]
            else Citation.CONF_HIGH
        )
        self.option["imp_citation_priv"] = (
            self.imp_values.citation_priv.get_privacy()
            if self.option["imp_citation"]
            else False
        )
        if self.option["imp_citation"]:
            citationpage = self.imp_values.citation_page.get_entry()
            if ("%Y" or "%m" or "%d" or "%H" or "%M" or "%S") in citationpage:
                citationpage = time.strftime(citationpage)
            self.option["imp_citation_page"] = citationpage
            citationattr = self.imp_values.citation_attr.get_entry()
            if ("%Y" or "%m" or "%d" or "%H" or "%M" or "%S") in citationattr:
                citationattr = time.strftime(citationattr)
            self.option["imp_citation_attr"] = citationattr
        else:
            self.option["imp_citation_page"] = ""
            self.option["imp_citation_attr"] = ""

        for obj in libprogen.TAGOBJECTS:
            if self.tagobj_values[obj].tag_obj.get_value():
                tagtext = self.tagobj_values[obj].tag_text.get_entry()
                if ("%Y" or "%m" or "%d" or "%H" or "%M" or "%S") in tagtext:
                    tagtext = time.strftime(tagtext)
                self.option["tag_{}".format(obj)] = tagtext
            else:
                self.option["tag_{}".format(obj)] = ""

        for obj in libprogen.PRIMOBJECTS:
            self.option["prim_{}".format(obj)] = self.primobj_values[obj].get_value()
        for obj in libprogen.OPTOBJECTS:
            self.option["opt_{}".format(obj)] = self.option_values[obj].get_value()

        self.fail = False  # Pro-Gen import proceed



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2002-2006  Donald N. Allingham
# Copyright (C) 2016       Nick Hall
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Rule that checks for a family that is an ancestor of a specified family.
"""
# -------------------------------------------------------------------------
#
# Standard python modules
#
# -------------------------------------------------------------------------
from __future__ import annotations

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from .. import Rule
from ....const import GRAMPS_LOCALE as glocale

# -------------------------------------------------------------------------
#
# Typing modules
#
# -------------------------------------------------------------------------
from typing import Set
from ....lib import Family
from ....db import Database


_ = glocale.translation.gettext


# -------------------------------------------------------------------------
#
# IsAncestorOf
#
# -------------------------------------------------------------------------
class IsAncestorOf(Rule):
    """
    Rule that checks for a family that is an ancestor of a specified family.
    """

    labels = [_("ID:"), _("Inclusive:")]
    name = _("Ancestor families of <family>")
    category = _("General filters")
    description = _("Matches ancestor families of the specified family")

    def prepare(self, db: Database, user):
        self.selected_handles: Set[str] = set()
        first = False if int(self.list[1]) else True
        root_family = db.get_family_from_gramps_id(self.list[0])
        self.init_list(db, root_family, first)

    def reset(self):
        self.selected_handles.clear()

    def apply_to_one(self, db: Database, family: Family) -> bool:
        return family.handle in self.selected_handles

    def init_list(self, db: Database, family: Family | None, first: bool) -> None:
        """
        Initialise family handle list.
        """
        if not family:
            return
        if family.handle in self.selected_handles:
            return
        if not first:
            self.selected_handles.add(family.handle)

        for parent_handle in [family.father_handle, family.mother_handle]:
            if parent_handle:
                parent = db.get_person_from_handle(parent_handle)
                family_handle = (
                    parent.parent_family_list[0]
                    if len(parent.parent_family_list) > 0
                    else None
                )
                if family_handle:
                    parent_family = db.get_family_from_handle(family_handle)
                    self.init_list(db, parent_family, False)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2006  Donald N. Allingham
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# -------------------------------------------------------------------------
#
# Python classes
#
# -------------------------------------------------------------------------
from gi.repository import GObject, GLib

# -------------------------------------------------------------------------
#
# Gramps classes
#
# -------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext
from gramps.gen.lib import RepoRef, Repository
from ...dbguielement import DbGUIElement
from ...selectors import SelectorFactory
from gramps.gen.errors import WindowActiveError
from ...ddtargets import DdTargets
from .reporefmodel import RepoRefModel
from .embeddedlist import EmbeddedList, TEXT_COL, MARKUP_COL, ICON_COL


# -------------------------------------------------------------------------
#
# RepoEmbedList
#
# -------------------------------------------------------------------------
class RepoEmbedList(EmbeddedList, DbGUIElement):
    _HANDLE_COL = 5
    _DND_TYPE = DdTargets.REPOREF
    _DND_EXTRA = DdTargets.REPO_LINK

    _MSG = {
        "add": _("Create and add a new repository"),
        "del": _("Remove the existing repository"),
        "edit": _("Edit the selected repository"),
        "share": _("Add an existing repository"),
        "up": _("Move the selected repository upwards"),
        "down": _("Move the selected repository downwards"),
    }

    # index = column in model. Value =
    #  (name, sortcol in model, width, markup/text, weigth_col
    _column_names = [
        (_("ID"), 0, 75, TEXT_COL, -1, None),
        (_("Name"), 1, 200, TEXT_COL, -1, None),
        (_("Call Number"), 2, 125, TEXT_COL, -1, None),
        (_("Type"), 3, 100, TEXT_COL, -1, None),
        (_("Private"), 4, 30, ICON_COL, -1, "gramps-lock"),
    ]

    def __init__(self, dbstate, uistate, track, obj, config_key):
        self.obj = obj
        EmbeddedList.__init__(
            self,
            dbstate,
            uistate,
            track,
            _("_Repositories"),
            RepoRefModel,
            config_key,
            share_button=True,
            move_buttons=True,
        )
        DbGUIElement.__init__(self, dbstate.db)
        self.callman.register_handles({"repository": [rref.ref for rref in self.obj]})

    def _connect_db_signals(self):
        """
        Implement base class DbGUIElement method
        """
        # note: repository-rebuild closes the editors, so no need to connect
        self.callman.register_callbacks(
            {
                "repository-delete": self.repo_delete,  # delete a repo we track
                "repository-update": self.repo_update,  # change a repo we track
            }
        )
        self.callman.connect_all(keys=["repository"])

    def get_icon_name(self):
        return "gramps-repository"

    def get_data(self):
        return self.obj

    def column_order(self):
        return ((1, 4), (1, 0), (1, 1), (1, 2), (1, 3))

    def handle_extra_type(self, objtype, obj):
        from .. import EditRepoRef

        try:
            ref = RepoRef()
            repo = self.dbstate.db.get_repository_from_handle(obj)
            EditRepoRef(
                self.dbstate, self.uistate, self.track, repo, ref, self.add_callback
            )
        except WindowActiveError:
            pass

    def add_button_clicked(self, obj):
        from .. import EditRepoRef

        ref = RepoRef()
        repo = Repository()
        try:
            EditRepoRef(
                self.dbstate, self.uistate, self.track, repo, ref, self.add_callback
            )
        except WindowActiveError:
            pass

    def share_button_clicked(self, obj):
        SelectRepository = SelectorFactory("Repository")

        sel = SelectRepository(self.dbstate, self.uistate, self.track)

        repo = sel.run()
        if repo:
            try:
                from .. import EditRepoRef

                ref = RepoRef()
                EditRepoRef(
                    self.dbstate, self.uistate, self.track, repo, ref, self.add_callback
                )

            except WindowActiveError:
                pass

    def add_callback(self, value):
        value[0].ref = value[1].handle
        data = self.get_data()
        data.append(value[0])
        self.callman.register_handles({"repository": [value[1].handle]})
        self.changed = True
        self.rebuild()
        GLib.idle_add(self.tree.scroll_to_cell, len(data) - 1)

    def edit_button_clicked(self, obj):
        ref = self.get_selected()
        if ref:
            repo = self.dbstate.db.get_repository_from_handle(ref.ref)
            try:
                from .. import EditRepoRef

                EditRepoRef(
                    self.dbstate,
                    self.uistate,
                    self.track,
                    repo,
                    ref,
                    self.edit_callback,
                )
            except WindowActiveError:
                from ...dialog import WarningDialog

                WarningDialog(
                    _("Cannot edit this reference"),
                    _(
                        "This repository reference cannot be edited at this "
                        "time. Either the associated repository is already "
                        "being edited or another repository reference that is "
                        "associated with the same repository is being edited."
                        "\n\nTo edit this repository reference, you need to "
                        "close the repository."
                    ),
                    parent=self.uistate.window,
                )

    def edit_callback(self, name):
        self.changed = True
        self.rebuild()

    def repo_delete(self, del_repo_handle_list):
        """
        Outside of this tab repo objects have been deleted. Check if tab
        and object must be changed.
        Note: delete of object will cause reference on database to be removed,
              so this method need not do this
        """
        rebuild = False
        ref_handles = [rref.ref for rref in self.obj]
        for handle in del_repo_handle_list:
            while 1:
                pos = None
                try:
                    pos = ref_handles.index(handle)
                except ValueError:
                    break

                if pos is not None:
                    # oeps, we need to remove this reference, and rebuild tab
                    del self.obj[pos]
                    del ref_handles[pos]
                    rebuild = True
        if rebuild:
            self.rebuild()

    def repo_update(self, upd_repo_handle_list):
        """
        Outside of this tab repo objects have been changed. Check if tab
        and object must be changed.
        """
        ref_handles = [rref.ref for rref in self.obj]
        for handle in upd_repo_handle_list:
            if handle in ref_handles:
                self.rebuild()
                break



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
# Copyright (C) 2010       Michiel D. Nauta
# Copyright (C) 2010       Jakim Friant
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Provide merge capabilities for persons.
"""

# -------------------------------------------------------------------------
#
# GTK/Gnome modules
#
# -------------------------------------------------------------------------
from gi.repository import Pango

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.sgettext
from gramps.gen.plug.report import utils
from gramps.gen.display.name import displayer as name_displayer
from gramps.gen.display.place import displayer as place_displayer
from gramps.gen.const import URL_MANUAL_SECT3
from ..display import display_help
from gramps.gen.datehandler import get_date
from gramps.gen.errors import MergeError
from ..dialog import ErrorDialog, WarningDialog
from ..managedwindow import ManagedWindow
from gramps.gen.merge import MergePersonQuery

# -------------------------------------------------------------------------
#
# Gramps constants
#
# -------------------------------------------------------------------------
WIKI_HELP_PAGE = URL_MANUAL_SECT3
WIKI_HELP_SEC = _("Merge_People", "manual")
_GLADE_FILE = "mergeperson.glade"

# Translators: needed for French, ignore otherwise
KEYVAL = _("%(key)s:\t%(value)s")

sex = (_("female"), _("male"), _("other"), _("unknown"))


def name_of(person):
    """Return string with name and ID of a person."""
    if not person:
        return ""
    return "%s [%s]" % (name_displayer.display(person), person.get_gramps_id())


# -------------------------------------------------------------------------
#
# MergePerson
#
# -------------------------------------------------------------------------
class MergePerson(ManagedWindow):
    """
    Displays a dialog box that allows the persons to be combined into one.
    """

    def __init__(
        self,
        dbstate,
        uistate,
        track,
        handle1,
        handle2,
        cb_update=None,
        expand_context_info=True,
    ):
        ManagedWindow.__init__(self, uistate, track, self.__class__)
        self.database = dbstate.db
        self.pr1 = self.database.get_person_from_handle(handle1)
        self.pr2 = self.database.get_person_from_handle(handle2)
        self.update = cb_update

        self.define_glade("mergeperson", _GLADE_FILE)
        self.set_window(
            self._gladeobj.toplevel, self.get_widget("person_title"), _("Merge People")
        )
        self.setup_configs("interface.merge-person", 700, 400)

        # Detailed selection widgets
        name1 = name_displayer.display_name(self.pr1.get_primary_name())
        name2 = name_displayer.display_name(self.pr2.get_primary_name())
        entry1 = self.get_widget("name1")
        entry2 = self.get_widget("name2")
        entry1.set_text(name1)
        entry2.set_text(name2)
        if entry1.get_text() == entry2.get_text():
            for widget_name in ("name1", "name2", "name_btn1", "name_btn2"):
                self.get_widget(widget_name).set_sensitive(False)

        entry1 = self.get_widget("gender1")
        entry2 = self.get_widget("gender2")
        entry1.set_text(sex[self.pr1.get_gender()])
        entry2.set_text(sex[self.pr2.get_gender()])
        if entry1.get_text() == entry2.get_text():
            for widget_name in ("gender1", "gender2", "gender_btn1", "gender_btn2"):
                self.get_widget(widget_name).set_sensitive(False)

        gramps1 = self.pr1.get_gramps_id()
        gramps2 = self.pr2.get_gramps_id()
        entry1 = self.get_widget("gramps1")
        entry2 = self.get_widget("gramps2")
        entry1.set_text(gramps1)
        entry2.set_text(gramps2)
        if entry1.get_text() == entry2.get_text():
            for widget_name in ("gramps1", "gramps2", "gramps_btn1", "gramps_btn2"):
                self.get_widget(widget_name).set_sensitive(False)

        # Main window widgets that determine which handle survives
        rbutton1 = self.get_widget("handle_btn1")
        rbutton_label1 = self.get_widget("label_handle_btn1")
        rbutton_label2 = self.get_widget("label_handle_btn2")
        rbutton_label1.set_label(name1 + " [" + gramps1 + "]")
        rbutton_label2.set_label(name2 + " [" + gramps2 + "]")
        rbutton1.connect("toggled", self.on_handle1_toggled)
        expander2 = self.get_widget("expander2")
        self.expander_handler = expander2.connect(
            "notify::expanded", self.cb_expander2_activated
        )
        expander2.set_expanded(expand_context_info)

        self.connect_button("person_help", self.cb_help)
        self.connect_button("person_ok", self.cb_merge)
        self.connect_button("person_cancel", self.close)
        self.show()

    def on_handle1_toggled(self, obj):
        """Preferred person changes"""
        if obj.get_active():
            self.get_widget("name_btn1").set_active(True)
            self.get_widget("gender_btn1").set_active(True)
            self.get_widget("gramps_btn1").set_active(True)
        else:
            self.get_widget("name_btn2").set_active(True)
            self.get_widget("gender_btn2").set_active(True)
            self.get_widget("gramps_btn2").set_active(True)

    def cb_expander2_activated(self, obj, param_spec):
        """Context Information expander is activated"""
        if obj.get_expanded():
            text1 = self.get_widget("text1")
            text2 = self.get_widget("text2")
            self.display(text1.get_buffer(), self.pr1)
            self.display(text2.get_buffer(), self.pr2)
            obj.disconnect(self.expander_handler)

    def add(self, tobj, tag, text):
        """Add text text to text buffer tobj with formatting tag."""
        text += "\n"
        tobj.insert_with_tags(tobj.get_end_iter(), text, tag)

    def display(self, tobj, person):
        """Fill text buffer tobj with detailed info on person person."""
        normal = tobj.create_tag()
        normal.set_property("indent", 10)
        normal.set_property("pixels-above-lines", 1)
        normal.set_property("pixels-below-lines", 1)
        indent = tobj.create_tag()
        indent.set_property("indent", 30)
        indent.set_property("pixels-above-lines", 1)
        indent.set_property("pixels-below-lines", 1)
        title = tobj.create_tag()
        title.set_property("weight", Pango.Weight.BOLD)
        title.set_property("scale", 1.2)
        self.add(tobj, title, name_displayer.display(person))
        self.add(
            tobj, normal, KEYVAL % {"key": _("ID"), "value": person.get_gramps_id()}
        )
        self.add(
            tobj,
            normal,
            KEYVAL % {"key": _("Gender"), "value": sex[person.get_gender()]},
        )
        bref = person.get_birth_ref()
        if bref:
            self.add(
                tobj,
                normal,
                KEYVAL % {"key": _("Birth"), "value": self.get_event_info(bref.ref)},
            )
        dref = person.get_death_ref()
        if dref:
            self.add(
                tobj,
                normal,
                KEYVAL % {"key": _("Death"), "value": self.get_event_info(dref.ref)},
            )

        nlist = person.get_alternate_names()
        if len(nlist) > 0:
            self.add(tobj, title, _("Alternate Names"))
            for name in nlist:
                self.add(tobj, normal, name_displayer.display_name(name))

        elist = person.get_event_ref_list()
        if len(elist) > 0:
            self.add(tobj, title, _("Events"))
            for event_ref in person.get_event_ref_list():
                event_handle = event_ref.ref
                role = event_ref.get_role()
                name = str(self.database.get_event_from_handle(event_handle).get_type())
                ev_info = self.get_event_info(event_handle)
                if role.is_primary():
                    self.add(tobj, normal, KEYVAL % {"key": name, "value": ev_info})
                else:
                    self.add(
                        tobj,
                        normal,  # Translators: needed for French
                        "%(name)s (%(role)s):\t%(info)s"
                        % {"name": name, "role": role, "info": ev_info},
                    )
        plist = person.get_parent_family_handle_list()

        if len(plist) > 0:
            self.add(tobj, title, _("Parents"))
            for fid in person.get_parent_family_handle_list():
                (fname, mname, gid) = self.get_parent_info(fid)
                self.add(tobj, normal, KEYVAL % {"key": _("Family ID"), "value": gid})
                if fname:
                    self.add(
                        tobj, indent, KEYVAL % {"key": _("Father"), "value": fname}
                    )
                if mname:
                    self.add(
                        tobj, indent, KEYVAL % {"key": _("Mother"), "value": mname}
                    )
        else:
            self.add(tobj, normal, _("No parents found"))

        self.add(tobj, title, _("Spouses"))
        slist = person.get_family_handle_list()
        if len(slist) > 0:
            for fid in slist:
                (fname, mname, pid) = self.get_parent_info(fid)
                family = self.database.get_family_from_handle(fid)
                self.add(tobj, normal, KEYVAL % {"key": _("Family ID"), "value": pid})
                spouse_id = utils.find_spouse(person, family)
                if spouse_id:
                    spouse = self.database.get_person_from_handle(spouse_id)
                    self.add(
                        tobj,
                        indent,
                        KEYVAL % {"key": _("Spouse"), "value": name_of(spouse)},
                    )
                relstr = str(family.get_relationship())
                self.add(tobj, indent, KEYVAL % {"key": _("Type"), "value": relstr})
                event = utils.find_marriage(self.database, family)
                if event:
                    m_info = self.get_event_info(event.get_handle())
                    self.add(
                        tobj, indent, KEYVAL % {"key": _("Marriage"), "value": m_info}
                    )
                for child_ref in family.get_child_ref_list():
                    child = self.database.get_person_from_handle(child_ref.ref)
                    self.add(
                        tobj,
                        indent,
                        KEYVAL % {"key": _("Child"), "value": name_of(child)},
                    )
        else:
            self.add(tobj, normal, _("No spouses or children found"))

        alist = person.get_address_list()
        if len(alist) > 0:
            self.add(tobj, title, _("Addresses"))
            for addr in alist:
                # TODO for Arabic, should the next line's comma be translated?
                location = ", ".join(
                    [
                        addr.get_street(),
                        addr.get_city(),
                        addr.get_state(),
                        addr.get_country(),
                        addr.get_postal_code(),
                        addr.get_phone(),
                    ]
                )
                self.add(tobj, normal, location.strip())

    def get_parent_info(self, fid):
        """Return tuple of father name, mother name and family ID"""
        family = self.database.get_family_from_handle(fid)
        father_id = family.get_father_handle()
        mother_id = family.get_mother_handle()
        if father_id:
            father = self.database.get_person_from_handle(father_id)
            fname = name_of(father)
        else:
            fname = ""
        if mother_id:
            mother = self.database.get_person_from_handle(mother_id)
            mname = name_of(mother)
        else:
            mname = ""
        return (fname, mname, family.get_gramps_id())

    def get_event_info(self, handle):
        """Return date and place of an event as string."""
        date = ""
        place = ""
        if handle:
            event = self.database.get_event_from_handle(handle)
            date = get_date(event)
            place = place_displayer.display_event(self.database, event)
            if date:
                return ("%s, %s" % (date, place)) if place else date
            else:
                return place or ""
        else:
            return ""

    def cb_help(self, obj):
        """Display the relevant portion of Gramps manual"""
        display_help(webpage=WIKI_HELP_PAGE, section=WIKI_HELP_SEC)

    def cb_merge(self, obj):
        """
        Perform the merge of the persons when the merge button is clicked.
        """
        self.uistate.set_busy_cursor(True)
        use_handle1 = self.get_widget("handle_btn1").get_active()
        if use_handle1:
            phoenix = self.pr1
            titanic = self.pr2
        else:
            phoenix = self.pr2
            titanic = self.pr1

        if self.get_widget("name_btn1").get_active() ^ use_handle1:
            swapname = phoenix.get_primary_name()
            phoenix.set_primary_name(titanic.get_primary_name())
            titanic.set_primary_name(swapname)
        if self.get_widget("gender_btn1").get_active() ^ use_handle1:
            phoenix.set_gender(titanic.get_gender())
        if self.get_widget("gramps_btn1").get_active() ^ use_handle1:
            swapid = phoenix.get_gramps_id()
            phoenix.set_gramps_id(titanic.get_gramps_id())
            titanic.set_gramps_id(swapid)

        try:
            query = MergePersonQuery(self.database, phoenix, titanic)
            family_merge_ok = query.execute()
            if not family_merge_ok:
                WarningDialog(
                    _("Warning"),
                    _(
                        "The persons have been merged.\nHowever, the families "
                        "for this merge were too complex to automatically "
                        "handle.  We recommend that you go to Relationships "
                        "view and see if additional manual merging of families "
                        "is necessary."
                    ),
                    parent=self.window,
                )
            # Add the selected handle to history so that when merge is complete,
            # phoenix is the selected row.
            self.uistate.set_active(phoenix.get_handle(), "Person")
        except MergeError as err:
            ErrorDialog(_("Cannot merge people"), str(err), parent=self.window)
        self.uistate.set_busy_cursor(False)
        self.close()
        if self.update:
            self.update()



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2002-2006  Donald N. Allingham
# Copyright (C) 2008       Gary Burton
# Copyright (C) 2010       Nick Hall
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# -------------------------------------------------------------------------
#
# Standard Python modules
#
# -------------------------------------------------------------------------
from ....const import GRAMPS_LOCALE as glocale

_ = glocale.translation.sgettext

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from .. import Rule
from ....lib import PlaceType
from ....utils.location import get_locations


# -------------------------------------------------------------------------
#
# Typing modules
#
# -------------------------------------------------------------------------
from ....lib import Place
from ....db import Database


# -------------------------------------------------------------------------
#
# HasPlace
#
# -------------------------------------------------------------------------
class HasPlace(Rule):
    """Rule that checks for a place with a particular value"""

    labels = [
        _("Title:"),
        _("Street:"),
        _("Locality:"),
        _("City:"),
        _("County:"),
        _("State:"),
        _("Country:"),
        _("ZIP/Postal Code:"),
        _("Church Parish:"),
    ]
    name = _("Places matching parameters")
    description = _("Matches places with particular parameters")
    category = _("General filters")
    allow_regex = True

    TYPE2FIELD = {
        PlaceType.STREET: 1,
        PlaceType.LOCALITY: 2,
        PlaceType.CITY: 3,
        PlaceType.COUNTY: 4,
        PlaceType.STATE: 5,
        PlaceType.COUNTRY: 6,
        PlaceType.PARISH: 8,
    }

    def apply_to_one(self, db: Database, place: Place) -> bool:
        if not self.match_substring(0, place.title):
            return False

        if not self.match_substring(7, place.code):
            return False

        # If no location data was given then we're done: match
        if not any(self.list[1:7] + [self.list[8]]):
            return True

        for location in get_locations(db, place):
            if self.check(location):
                return True

        return False

    def check(self, location):
        """
        Check each location for a match.
        """
        for place_type, field in self.TYPE2FIELD.items():
            name_list = location.get(place_type, [""])
            if not self.match_name(field, name_list):
                return False
        return True

    def match_name(self, field, name_list):
        """
        Match any name in a list of names.
        """
        for name in name_list:
            if self.match_substring(field, name):
                return True
        return False



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2002-2006  Donald N. Allingham
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Rule matching people who are witnesses in any event.
"""

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from ....const import GRAMPS_LOCALE as glocale
from ....lib.eventroletype import EventRoleType
from ....lib.eventtype import EventType
from .. import Rule

# -------------------------------------------------------------------------
#
# Typing modules
#
# -------------------------------------------------------------------------
from ....lib import Person
from ....db import Database


_ = glocale.translation.gettext


# -------------------------------------------------------------------------
#
# IsWitness
#
# -------------------------------------------------------------------------
class IsWitness(Rule):
    """
    Rule matching people who are witnesses in any event.
    """

    labels = [_("Event type:")]
    name = _("Witnesses")
    description = _("Matches people who are witnesses in any event")
    category = _("Event filters")

    def __init__(self, arg, use_regex=False, use_case=False):
        super().__init__(arg, use_regex, use_case)
        self.event_type = None

    def prepare(self, db: Database, user):
        """
        Prepare the rule. Things only want to do once.
        """
        if self.list[0]:
            self.event_type = EventType()
            self.event_type.set_from_xml_str(self.list[0])

    def apply_to_one(self, db: Database, obj: Person) -> bool:
        """
        Apply the rule. Return True on a match.
        """
        for event_ref in obj.event_ref_list:
            if event_ref.role.value == EventRoleType.WITNESS:
                # This is the witness.
                # If event type was given, then check it.
                if self.event_type:
                    event = db.get_event_from_handle(event_ref.ref)
                    if event.type == self.event_type:
                        return True
                else:
                    # event type was not specified, we're returning a match
                    return True
        return False



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2002-2006  Donald N. Allingham
# Copyright (C) 2011       Tim G L Lyons
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# -------------------------------------------------------------------------
#
# Standard Python modules
#
# -------------------------------------------------------------------------
from ....const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext
import logging

LOG = logging.getLogger(".citationfilter")

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from ....utils.db import get_source_and_citation_referents
from .. import Rule


# -------------------------------------------------------------------------
#
# Typing modules
#
# -------------------------------------------------------------------------
from typing import Set
from ....lib import Person
from ....db import Database
from ....types import (
    PersonHandle,
    EventHandle,
    PlaceHandle,
    MediaHandle,
    FamilyHandle,
)


# -------------------------------------------------------------------------
# "HasTextMatchingSubstringOf"
# -------------------------------------------------------------------------
class HasTextMatchingSubstringOf(Rule):
    """Rule that checks for string matches in any textual information"""

    labels = [_("Substring:"), _("Case sensitive:")]
    name = _("People with records containing <substring>")
    description = _("Matches people whose records contain text " "matching a substring")
    category = _("General filters")
    allow_regex = True

    def prepare(self, db: Database, user):
        self.db = db
        self.person_map: Set[str] = set()
        self.event_map: Set[str] = set()
        self.source_map: Set[str] = set()
        self.repo_map: Set[str] = set()
        self.family_map: Set[str] = set()
        self.place_map: Set[str] = set()
        self.media_map: Set[str] = set()
        try:
            if int(self.list[1]):
                self.case_sensitive = True
            else:
                self.case_sensitive = False
        except IndexError:
            self.case_sensitive = False
        self.cache_repos()
        self.cache_sources()

    def reset(self):
        self.person_map.clear()
        self.event_map.clear()
        self.source_map.clear()
        self.repo_map.clear()
        self.family_map.clear()
        self.place_map.clear()
        self.media_map.clear()

    def apply_to_one(self, db: Database, person: Person) -> bool:
        if person.handle in self.person_map:  # Cached by matching Source?
            return True
        if self.match_object(person):  # first match the person itself
            return True

        # Look for matching events
        if any(self.search_event(event_ref.ref) for event_ref in person.event_ref_list):
            return True

        # Look for matching families
        if any(
            self.search_family(family_handle) for family_handle in person.family_list
        ):
            return True

        # Look for matching media objects
        if any(self.search_media(media_ref.ref) for media_ref in person.media_list):
            return True
        return False

    def search_family(self, family_handle: FamilyHandle):
        if not family_handle:
            return False
        # search inside the family and cache the result to not search a family twice
        if family_handle not in self.family_map:
            match = 0
            family = self.db.get_family_from_handle(family_handle)
            if self.match_object(family):
                match = 1
            else:
                if any(
                    self.search_event(event_ref.ref)
                    for event_ref in family.event_ref_list
                ):
                    match = 1
                if any(
                    self.search_media(media_ref.ref) for media_ref in family.media_list
                ):
                    return True
            if match:
                self.family_map.add(family_handle)
        return family_handle in self.family_map

    def search_event(self, event_handle: EventHandle):
        if not event_handle:
            return False
        # search inside the event and cache the result (event sharing)
        if not event_handle in self.event_map:
            match = 0
            event = self.db.get_event_from_handle(event_handle)
            if self.match_object(event):
                match = 1
            elif event:
                place_handle = event.place
                if place_handle and self.search_place(place_handle):
                    match = 1
                if any(
                    self.search_media(media_ref.ref) for media_ref in event.media_list
                ):
                    return True
            if match:
                self.event_map.add(event_handle)
        return event_handle in self.event_map

    def search_place(self, place_handle: PlaceHandle) -> bool:
        if not place_handle:
            return False
        # search inside the place and cache the result
        if place_handle not in self.place_map:
            place = self.db.get_place_from_handle(place_handle)
            if self.match_object(place):
                self.place_map.add(place_handle)
        return place_handle in self.place_map

    def search_media(self, media_handle: MediaHandle) -> bool:
        if not media_handle:
            return False
        # search inside the media object and cache the result
        if media_handle not in self.media_map:
            media = self.db.get_media_from_handle(media_handle)
            if self.match_object(media):
                self.media_map.add(media_handle)
        return media_handle in self.media_map

    def cache_repos(self):
        # search all matching repositories
        self.repo_map.update(
            repo.handle
            for repo in self.db.iter_repositories()
            if repo and self.match_object(repo)
        )

    def cache_sources(self):
        # search all sources and match all referents of a matching source
        for source in self.db.iter_sources():
            match = self.match_object(source)
            LOG.debug(
                "cache_sources match %s string %s source %s"
                % (match, self.list[0], source.gramps_id)
            )
            if not match:
                if any(reporef.ref in self.repo_map for reporef in source.reporef_list):
                    match = True
                    LOG.debug(
                        "cache_sources repomatch %s string %s source %s"
                        % (match, self.list[0], source.gramps_id)
                    )
            (
                citation_list,
                citation_referents_list,
            ) = get_source_and_citation_referents(source.handle, self.db)
            LOG.debug("the_lists %s %s" % (citation_list, citation_referents_list))
            for citation_handle, refs in citation_referents_list:
                citation = self.db.get_citation_from_handle(citation_handle)
                LOG.debug(
                    "cache_sources match %s matchcitation %s string %s"
                    " source %s citation %s"
                    % (
                        match,
                        self.match_object(citation),
                        self.list[0],
                        source.gramps_id,
                        citation.gramps_id,
                    )
                )
                if match or self.match_object(citation):
                    # Update the maps to reflect the reference
                    (
                        person_list,
                        family_list,
                        event_list,
                        place_list,
                        source_list,
                        media_list,
                        repo_list,
                    ) = refs
                    self.person_map.update(person_list)
                    self.family_map.update(family_list)
                    self.event_map.update(event_list)
                    self.place_map.update(place_list)
                    self.media_map.update(media_list)
                    self.repo_map.update(repo_list)

    def match_object(self, obj):
        if not obj:
            return False
        if self.use_regex:
            return obj.matches_regexp(self.list[0], self.case_sensitive)
        return obj.matches_string(self.list[0], self.case_sensitive)



# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2011 Nick Hall
# Copyright (C) 2011 Tim G L Lyons
# Copyright (C) 2020 Matthias Kemmer
# Copyright (C) 2024-2025 Steve Youngs
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# ------------------------------------------------------------------------
#
# Gtk
#
# ------------------------------------------------------------------------
from gi.repository import Gtk

# ------------------------------------------------------------------------
#
# Gramps modules
#
# ------------------------------------------------------------------------
from gramps.gen.plug import Gramplet
from gramps.gui.editors import EditNote
from gramps.gui.widgets.styledtexteditor import StyledTextEditor
from gramps.gui.widgets import SimpleButton
from gramps.gen.errors import WindowActiveError
from gramps.gen.lib import StyledText
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext


class Notes(Gramplet):
    """
    Displays the notes for an object.
    """

    def init(self):
        self.gui.WIDGET = self.build_gui()
        self.gui.get_container_widget().remove(self.gui.textview)
        self.gui.get_container_widget().add(self.gui.WIDGET)
        self.gui.WIDGET.show()

    def build_gui(self):
        """
        Build the GUI interface.
        """
        top = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)

        hbox = Gtk.Box()
        self.left = SimpleButton("go-previous", self.left_clicked)
        self.left.set_sensitive(False)
        hbox.pack_start(self.left, False, False, 0)
        self.right = SimpleButton("go-next", self.right_clicked)
        self.right.set_sensitive(False)
        hbox.pack_start(self.right, False, False, 0)
        self.edit = SimpleButton("gtk-edit", self.edit_clicked)
        self.edit.set_sensitive(False)
        hbox.pack_start(self.edit, False, False, 0)
        self.page = Gtk.Label()
        self.page.set_halign(Gtk.Align.START)
        hbox.pack_start(self.page, True, True, 10)
        self.ntype = Gtk.Label()
        hbox.pack_start(self.ntype, False, False, 10)

        scrolledwindow = Gtk.ScrolledWindow()
        scrolledwindow.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)
        self.texteditor = StyledTextEditor()
        self.texteditor.set_editable(False)
        self.texteditor.set_wrap_mode(Gtk.WrapMode.WORD)
        scrolledwindow.add(self.texteditor)

        top.pack_start(hbox, False, False, 0)
        top.pack_start(scrolledwindow, True, True, 0)
        top.show_all()
        return top

    def clear_text(self):
        self.left.set_sensitive(False)
        self.right.set_sensitive(False)
        self.edit.set_sensitive(False)
        self.texteditor.set_text(StyledText())
        self.page.set_text("")
        self.current = 0

    def display_note(self):
        """
        Display the current note.
        """
        note_handle = self.note_list[self.current]
        note = self.dbstate.db.get_note_from_handle(note_handle)
        self.texteditor.set_text(note.get_styledtext())
        self.ntype.set_text(str(note.get_type()))
        self.page.set_text(
            _("%(current)d of %(total)d")
            % {"current": self.current + 1, "total": len(self.note_list)}
        )

    def left_clicked(self, button):
        """
        Display the previous note.
        """
        if self.current > 0:
            self.current -= 1
            self.right.set_sensitive(True)
            if self.current == 0:
                self.left.set_sensitive(False)
            self.display_note()

    def right_clicked(self, button):
        """
        Display the next note.
        """
        if self.current < len(self.note_list) - 1:
            self.current += 1
            self.left.set_sensitive(True)
            if self.current == len(self.note_list) - 1:
                self.right.set_sensitive(False)
            self.display_note()

    def edit_clicked(self, button):
        pass

    def get_has_data(self, obj):
        """
        Return True if the gramplet has data, else return False.
        """
        if obj is None:
            return False
        if obj.get_note_list():
            return True
        return False


class NotesOf(Notes):
    def init(self, object_class):
        super().init()
        self.object_class = object_class

    def db_changed(self):
        self.connect(
            self.dbstate.db, "%s-update" % self.object_class.lower(), self.update
        )
        self.connect(
            self.dbstate.db, "%s-delete" % self.object_class.lower(), self.update
        )
        self.connect(self.dbstate.db, "note-add", self.update)
        self.connect(self.dbstate.db, "note-update", self.update)
        self.connect(self.dbstate.db, "note-delete", self.update)

    def active_changed(self, handle):
        self.update()
        self.connect_signal(self.object_class, self.update)

    def get_notes(self, obj):
        """
        Get the note list for the current object.
        """
        self.left.set_sensitive(False)
        self.right.set_sensitive(False)
        self.texteditor.set_text(StyledText())
        self.note_list = obj.get_note_list()
        self.page.set_text("")
        self.ntype.set_text("")
        self.edit.set_sensitive(len(self.note_list) > 0)
        if len(self.note_list) > 0:
            self.set_has_data(True)
            if len(self.note_list) > 1:
                self.right.set_sensitive(True)
            self.current = 0
            self.display_note()
        else:
            self.set_has_data(False)

    def update_has_data(self):
        active_handle = self.get_active(self.object_class)
        if active_handle:
            active = self.dbstate.db.method("get_%s_from_handle", self.object_class)(
                active_handle
            )
            self.set_has_data(self.get_has_data(active))
        else:
            self.set_has_data(False)

    def edit_clicked(self, button):
        """
        Get the selected Note instance and call the EditNote editor with the
        note.

        Called when the Edit button is clicked.
        If the window already exists (WindowActiveError), we ignore it.
        This prevents the dialog from coming up twice on the same object.
        """
        note_handle = self.note_list[self.current]
        note = self.dbstate.db.get_note_from_handle(note_handle)
        try:
            EditNote(self.gui.dbstate, self.gui.uistate, [], note)
        except WindowActiveError:
            pass

    def main(self):
        self.clear_text()
        active_handle = self.get_active(self.object_class)
        if active_handle:
            active = self.dbstate.db.method("get_%s_from_handle", self.object_class)(
                active_handle
            )
            if active:
                self.get_notes(active)
            else:
                self.set_has_data(False)
        else:
            self.set_has_data(False)


class PersonNotes(NotesOf):
    """
    Displays the notes for a person.
    """

    def init(self):
        super().init("Person")

    def db_changed(self):
        self.connect(self.dbstate.db, "person-update", self.update)
        # superclass will call active_changed if the active Person changes
        self.connect(self.dbstate.db, "note-add", self.update)
        self.connect(self.dbstate.db, "note-update", self.update)
        self.connect(self.dbstate.db, "note-delete", self.update)

    def active_changed(self, handle):
        self.update()


class EventNotes(NotesOf):
    """
    Displays the notes for an event.
    """

    def init(self):
        super().init("Event")


class FamilyNotes(NotesOf):
    """
    Displays the notes for a family.
    """

    def init(self):
        super().init("Family")


class PlaceNotes(NotesOf):
    """
    Displays the notes for a place.
    """

    def init(self):
        super().init("Place")


class SourceNotes(NotesOf):
    """
    Displays the notes for a source.
    """

    def init(self):
        super().init("Source")


class CitationNotes(NotesOf):
    """
    Displays the notes for a Citation.
    """

    def init(self):
        super().init("Citation")


class RepositoryNotes(NotesOf):
    """
    Displays the notes for a repository.
    """

    def init(self):
        super().init("Repository")


class MediaNotes(NotesOf):
    """
    Displays the notes for a media object.
    """

    def init(self):
        super().init("Media")


class NoteNotes(Notes):
    """
    Display a single note in NoteView.
    """

    def db_changed(self):
        self.connect(self.dbstate.db, "note-update", self.update)
        self.connect(self.dbstate.db, "note-delete", self.update)
        self.connect_signal("Note", self.update)

    def edit_clicked(self, button):
        """
        Get the selected Note instance and call the EditNote editor with the
        note.

        Called when the Edit button is clicked.
        If the window already exists (WindowActiveError), we ignore it.
        This prevents the dialog from coming up twice on the same object.
        """
        active_handle = self.get_active("Note")
        note = self.dbstate.db.get_note_from_handle(active_handle)
        try:
            EditNote(self.gui.dbstate, self.gui.uistate, [], note)
        except WindowActiveError:
            pass

    def main(self):
        self.clear_text()
        active_handle = self.get_active("Note")
        if active_handle:
            active = self.dbstate.db.get_note_from_handle(active_handle)
            if active:
                self.edit.set_sensitive(True)
                self.texteditor.set_text(active.get_styledtext())



# -*- coding: utf-8 -*-
#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2013  Vassilii Khachaturov
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

import unittest
from unittest.mock import Mock


class LexGettextTest(unittest.TestCase):
    SRC_WORD = "Inflect-me"
    CONTEXT = "how-to-use-lexgettext"
    MSGID = "|" + SRC_WORD

    def setUp(self):
        from ..grampslocale import GrampsTranslations
        from ..grampslocale import GrampsLocale as Loc

        self.trans = GrampsTranslations()

    def setup_sgettext_mock(self, msgval_expected):
        mock = Mock(return_value=msgval_expected)
        self.trans.sgettext = mock

    def tearDown(self):
        try:
            self.trans.sgettext.assert_called_once_with(self.MSGID, self.CONTEXT)
        except AttributeError as e:
            print("Apparently the test has never set up the mock: ", e)

    def testSrcWordOnlyIfNoTranslation(self):
        self.setup_sgettext_mock(self.SRC_WORD)
        result = self.trans.lexgettext(self.MSGID, self.CONTEXT)
        self.assertEqual(result, self.SRC_WORD)

    def test3InflectionsExtractableByNameThroughForm(self):
        translated = "n=TargetNom|g=TargetGen|d=TargetDat"
        self.setup_sgettext_mock(translated)
        lex = self.trans.lexgettext(self.MSGID, self.CONTEXT)
        formatted = "{lex.forms[n]},{lex.forms[g]},{lex.forms[d]}".format(lex=lex)
        self.assertEqual(formatted, "TargetNom,TargetGen,TargetDat")

    def testFirstLexemeFormExtractableAsDefaultString(self):
        translated = "def=Default|v1=Option1|a=AnotherOption"
        self.setup_sgettext_mock(translated)
        lex = self.trans.lexgettext(self.MSGID, self.CONTEXT)
        formatted = "{}".format(lex)
        self.assertEqual(formatted, "Default")


class LexemeTest(unittest.TestCase):
    def setUp(self):
        from ..grampstranslation import Lexeme

        self.lex = Lexeme((("a", "aaa"), ("b", "bbb"), ("c", "ccc")))
        self.zlex = Lexeme({"z": "zzz"})
        self.elex = Lexeme({})

    def testIsHashable(self):
        hash(self.lex)  # throws if not hashable

    # test delegation to an arbitrary string method pulled in from unicode
    def testDefaultStringStartsWithAA(self):
        self.assertTrue(
            self.lex.startswith("aa"),
            msg="default string: {} dict: {}".format(self.lex, self.lex.__dict__),
        )

    def testCanConcatenateStringAndLexeme(self):
        moo = "moo"
        self.assertEqual(moo + self.lex, "mooaaa")

    def testCanConcatenateStringAndLexemeInPlace(self):
        moo = "moo"
        moo += self.lex
        self.assertEqual(moo, "mooaaa")

    def testCanConcatenateLexemeAndStringInPlace(self):
        moo = "moo"
        self.lex += moo
        self.assertEqual(self.lex, "aaamoo")

    def testCanConcatenateTwoLexemes(self):
        aaazzz = self.lex + self.zlex
        self.assertEqual(aaazzz, "aaazzz")

    def testCanJoinTwoLexemes(self):
        aaa_zzz = "_".join([self.lex, self.zlex])
        self.assertEqual(aaa_zzz, "aaa_zzz")

    def testEmptyIterableLikeEmptyString(self):
        self.assertEqual(self.elex, "")


class AddonTranslatorTest(unittest.TestCase):
    """
    Tests for the addon translator.
    """

    def setUp(self):
        from ...const import GRAMPS_LOCALE as glocale

        self._ = glocale.get_addon_translator(__file__, languages=["fr"]).gettext

    def testAddon(self):
        # A string in the "addon" domain.
        self.assertEqual(self._("Test Message"), "Message d'essai")

    def testGramps(self):
        # A string in the "gramps" domain.
        self.assertEqual(self._("United States of America"), "États-Unis")

    def testUntranslated(self):
        # An untranslated string.
        self.assertEqual(self._("Untranslated string"), "Untranslated string")


if __name__ == "__main__":
    unittest.main()



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
# Copyright (C) 2016       Tim G L Lyons
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# test/LosHawlos_dbtest.py

"""
Simple test for bsdbd. Minimal adaptations made (2016) from original version to
make the code compile and run under nosetests (unittest). This runs from the
root github directory with a command line like:
nosetests gramps test
where gramps is the directory containing most of the Gramps code, and this file
is in the directory 'test'
"""
import os
import tempfile
import shutil
import sys

from gramps.gen.dbstate import DbState
from gramps.cli.clidbman import CLIDbManager
from gramps.gen.db.base import DbTxn
from gramps.gen.db.utils import make_database
from gramps.gen.errors import DbError, HandleError
import gramps.gen.const as const
import gramps.gen.lib as RelLib

print(sys.path)

tran = None


def dummy_callback(dummy):
    pass


def add_source(db, title, commit=True, fail=False):
    global tran
    if tran is None:
        tran = db.transaction_begin(DbTxn("add source", db))
    db.disable_signals()
    s = RelLib.Source()
    db.add_source(s, tran)
    s.set_title(title)
    if fail:
        return  # Fail here
    db.commit_source(s, tran)
    db.enable_signals()
    if commit:
        db.transaction_commit(tran)
        tran = None


def add_person(db, firstname, lastname, commit=True, fail=False):
    global tran
    if tran is None:
        tran = db.transaction_begin(DbTxn("add person", db))
    db.disable_signals()
    p = RelLib.Person()
    db.add_person(p, tran)
    n = RelLib.Name()
    n.set_first_name(firstname)
    s = RelLib.Surname()
    s.set_surname(lastname)
    n.add_surname(s)
    p.set_primary_name(n)
    if fail:
        return  # Fail here
    db.commit_person(p, tran)
    db.enable_signals()
    if commit:
        db.transaction_commit(tran)
        tran = None


def print_db_content(db):
    for h in db.get_person_handles():
        print("DB contains: person %s" % h)
    for h in db.get_source_handles():
        print("DB contains: source %s" % h)


tmpdir = tempfile.mkdtemp()
try:
    filename1 = os.path.join(tmpdir, "test1.grdb")
    filename2 = os.path.join(tmpdir, "test2.grdb")
    print("\nUsing Database file: %s" % filename1)
    dbstate = DbState()
    dbman = CLIDbManager(dbstate)
    dirpath, name = dbman.create_new_db_cli(filename1, dbid="bsddb")
    db = make_database("bsddb")
    db.load(dirpath, None)
    print("Add person 1")
    add_person(db, "Anton", "Albers", True, False)
    print("Add source")
    add_source(db, "A short test", True, False)
    print("Add person 2 without commit")
    add_person(db, "Bernd", "Beta", False, False)
    print("Add source")
    add_source(db, "A short test", True, False)
    print("Add person 3")
    add_person(db, "Chris", "Connor", True, False)
    print_db_content(db)
    print("Closing Database file: %s" % filename1)
    db.close()
    tran = None

    print("\nUsing Database file: %s" % filename1)
    dbstate = DbState()
    dbman = CLIDbManager(dbstate)
    dirpath, name = dbman.create_new_db_cli(filename1, dbid="bsddb")
    db = make_database("bsddb")
    db.load(dirpath, None)
    print("Add person 4")
    add_person(db, "Felix", "Fowler", True, False)
    print("Add person 4")
    add_person(db, "Felix", "Fowler", False, False)
    print_db_content(db)
    print("Closing Database file: %s" % filename1)
    db.close()
    tran = None

    print("\nUsing Database file: %s" % filename2)
    dbstate = DbState()
    dbman = CLIDbManager(dbstate)
    dirpath, name = dbman.create_new_db_cli(filename2, dbid="bsddb")
    db = make_database("bsddb")
    db.load(dirpath, None)

    print("Add source")
    add_source(db, "A short test", False, False)
    # actually, adding a second source while the first transaction is not
    # committed will just add the second source to the first transaction, so
    # nothing special will fail.
    print("Add source 2 will fail; but I don't see why it should")
    add_source(db, "Bang bang bang", True, True)

    print_db_content(db)
    print("Closing Database file: %s" % filename2)
    db.close()
finally:
    print("Exit. Cleaning up.")
    shutil.rmtree(tmpdir)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
# Copyright (C) 2010       Michiel D. Nauta
# Copyright (C) 2010       Jakim Friant
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Provide merge capabilities for persons.
"""

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from ..db import DbTxn
from ..const import GRAMPS_LOCALE as glocale
from ..errors import MergeError

_ = glocale.translation.sgettext


# -------------------------------------------------------------------------
#
# MergePersonQuery
#
# -------------------------------------------------------------------------
class MergePersonQuery:
    """
    Create database query to merge two persons.
    """

    def __init__(self, database, phoenix, titanic):
        self.database = database
        self.phoenix = phoenix
        self.titanic = titanic
        if self.check_for_spouse(self.phoenix, self.titanic):
            raise MergeError(
                _(
                    "Spouses cannot be merged. To merge these "
                    "people, you must first break the relationship"
                    " between them."
                )
            )
        if self.check_for_child(self.phoenix, self.titanic):
            raise MergeError(
                _(
                    "A parent and child cannot be merged. To merge "
                    "these people, you must first break the relatio"
                    "nship between them."
                )
            )

    def check_for_spouse(self, person1, person2):
        """Return if person1 and person2 are spouses of eachother."""
        fs1 = set(person1.get_family_handle_list())
        fs2 = set(person2.get_family_handle_list())
        return len(fs1.intersection(fs2)) != 0

    def check_for_child(self, person1, person2):
        """Return if person1 and person2 have a child-parent relationship."""
        fs1 = set(person1.get_family_handle_list())
        fp1 = set(person1.get_parent_family_handle_list())
        fs2 = set(person2.get_family_handle_list())
        fp2 = set(person2.get_parent_family_handle_list())
        return len(fs1.intersection(fp2)) != 0 or len(fs2.intersection(fp1))

    def merge_families(self, main_family_handle, family, trans):
        """
        Merge content of family into the family with handle main_family_handle.
        """
        new_handle = self.phoenix.get_handle() if self.phoenix else None
        old_handle = self.titanic.get_handle() if self.titanic else None
        family_handle = family.get_handle()
        main_family = self.database.get_family_from_handle(main_family_handle)
        main_family.merge(family)
        for childref in family.get_child_ref_list():
            child = self.database.get_person_from_handle(
                childref.get_reference_handle()
            )
            if main_family_handle in child.parent_family_list:
                child.remove_handle_references("Family", [family_handle])
            else:
                child.replace_handle_reference(
                    "Family", family_handle, main_family_handle
                )
            self.database.commit_person(child, trans)
        if self.phoenix:
            self.phoenix.remove_family_handle(family_handle)
            self.database.commit_person(self.phoenix, trans)
        family_father_handle = family.get_father_handle()
        spouse_handle = (
            family.get_mother_handle()
            if new_handle == family_father_handle
            else family_father_handle
        )
        if spouse_handle:
            spouse = self.database.get_person_from_handle(spouse_handle)
            if spouse:
                spouse.remove_family_handle(family_handle)
                self.database.commit_person(spouse, trans)
        # replace the family in lds ordinances
        for ref_obj, ref_handle in self.database.find_backlink_handles(
            family_handle, ["Person", "Note"]
        ):
            if ref_handle == old_handle:
                continue
            obj = self.database.method("get_%s_from_handle", ref_obj)(ref_handle)
            assert obj.has_handle_reference("Family", family_handle)
            obj.replace_handle_reference("Family", family_handle, main_family_handle)
            self.database.method("commit_%s", ref_obj)(obj, trans)
        self.database.remove_family(family_handle, trans)
        self.database.commit_family(main_family, trans)

    def execute(self, family_merger=True, trans=None):
        """
        Merges two persons into a single person.
        """
        if trans is None:
            with DbTxn(_("Merge Person"), self.database) as trans:
                return self.__execute(family_merger, trans)
        else:
            return self.__execute(family_merger, trans)

    def __execute(self, family_merger, trans):
        """
        Merges two persons into a single person; trans is compulsory.
        Returns True if all went well, False if we did not complete a full
                family merge, because of complex families
        """
        new_handle = self.phoenix.get_handle()
        old_handle = self.titanic.get_handle()

        self.phoenix.merge(self.titanic)
        self.database.commit_person(self.phoenix, trans)

        for ref_obj, handle in self.database.find_backlink_handles(
            old_handle, ["Person", "Note"]
        ):
            obj = self.database.method("get_%s_from_handle", ref_obj)(handle)
            assert obj.has_handle_reference("Person", old_handle)
            obj.replace_handle_reference("Person", old_handle, new_handle)
            if handle != old_handle:
                self.database.method("commit_%s", ref_obj)(obj, trans)

        for family_handle in self.phoenix.get_parent_family_handle_list():
            family = self.database.get_family_from_handle(family_handle)
            if family.has_handle_reference("Person", old_handle):
                family.replace_handle_reference("Person", old_handle, new_handle)
                self.database.commit_family(family, trans)

        family_merge_guard = False
        parent_list = []
        parent_list_orig = []
        family_merge_ok = True
        family_handle_list = self.phoenix.get_family_handle_list()[:]
        for family_handle in family_handle_list:
            family = self.database.get_family_from_handle(family_handle)
            parents = (family.get_father_handle(), family.get_mother_handle())
            parent_list_orig.append(parents)
            if family.has_handle_reference("Person", old_handle):
                family.replace_handle_reference("Person", old_handle, new_handle)
                if family_merger and parent_list_orig.count(parents) > 1:
                    # A person with multiple relations with the same spouse is
                    # about to be merged. This is beyond the capabilities of
                    # the merge routine. The family merge is skipped.
                    family_merge_ok = False
                parents = (family.get_father_handle(), family.get_mother_handle())
                # prune means merging families in this case.
                if family_merger and parents in parent_list and family_merge_ok:
                    # also merge when father_handle or mother_handle == None!
                    if family_merge_guard:
                        # Multiple families get merged.
                        # This is unusual, the merge is skipped.
                        family_merge_ok = False
                    else:
                        idx = parent_list.index(parents)
                        main_family_handle = family_handle_list[idx]
                        self.merge_families(main_family_handle, family, trans)
                        family_merge_guard = True
                        continue
                self.database.commit_family(family, trans)
            parent_list.append(parents)

        hp_hdl = self.database.get_default_handle()
        if (
            hp_hdl in (self.phoenix.get_handle(), self.titanic.get_handle())
            and hp_hdl != self.phoenix.get_handle()
        ):
            self.database.set_default_person_handle(self.phoenix.get_handle())

        self.database.remove_person(old_handle, trans)
        return family_merge_ok



# encoding:utf-8
#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2009 Benny Malengier
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
from gramps.gen.plug._pluginreg import register, STABLE, GENERAL
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext

MODULE_VERSION = "6.0"

# ------------------------------------------------------------------------
#
# libcairo
#
# ------------------------------------------------------------------------
register(
    GENERAL,
    id="libcairodoc",
    name="Cairodoc lib",
    description=_("Provides a library for using Cairo to " "generate documents."),
    version="1.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="libcairodoc.py",
    authors=["The Gramps project"],
    authors_email=["http://gramps-project.org"],
    # load_on_reg = True
)

# ------------------------------------------------------------------------
#
# libgedcom
#
# ------------------------------------------------------------------------
register(
    GENERAL,
    id="libgedcom",
    name="GEDCOM library",
    description=_("Provides GEDCOM processing functionality"),
    version="1.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="libgedcom.py",
    authors=["The Gramps project"],
    authors_email=["http://gramps-project.org"],
)

# ------------------------------------------------------------------------
#
# librecurse
#
# ------------------------------------------------------------------------
register(
    GENERAL,
    id="librecurse",
    name="Recursive lib",
    description=_("Provides recursive routines for reports"),
    version="1.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="librecurse.py",
    authors=["The Gramps project"],
    authors_email=["http://gramps-project.org"],
)

# ------------------------------------------------------------------------
#
# libgrampsxml
#
# ------------------------------------------------------------------------
register(
    GENERAL,
    id="libgrampsxml",
    name="Grampsxml lib",
    description=_("Provides common functionality for Gramps XML " "import/export."),
    version="1.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="libgrampsxml.py",
    authors=["The Gramps project"],
    authors_email=["http://gramps-project.org"],
    # load_on_reg = True
)

# ------------------------------------------------------------------------
#
# libholiday
#
# ------------------------------------------------------------------------
register(
    GENERAL,
    id="libholiday",
    name="holiday lib",
    description=_("Provides holiday information for different countries."),
    version="1.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="libholiday.py",
    authors=["The Gramps project"],
    authors_email=["http://gramps-project.org"],
    # load_on_reg = True
)

# ------------------------------------------------------------------------
#
# llibhtmlbackend
#
# ------------------------------------------------------------------------
register(
    GENERAL,
    id="libhtmlbackend",
    name="htmlbackend lib",
    description=_("Manages a HTML file implementing DocBackend."),
    version="1.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="libhtmlbackend.py",
    authors=["The Gramps project"],
    authors_email=["http://gramps-project.org"],
    # load_on_reg = True
)

# ------------------------------------------------------------------------
#
# libhtmlconst
#
# ------------------------------------------------------------------------
register(
    GENERAL,
    id="libhtmlconst",
    name="htmlconst lib",
    description=_("Common constants for html files."),
    version="1.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="libhtmlconst.py",
    authors=["The Gramps project"],
    authors_email=["http://gramps-project.org"],
    # load_on_reg = True
)

# ------------------------------------------------------------------------
#
# libhtml
#
# ------------------------------------------------------------------------
register(
    GENERAL,
    id="libhtml",
    name="html lib",
    description=_("Manages an HTML DOM tree."),
    version="1.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="libhtml.py",
    authors=["Gerald Britton"],
    authors_email=["gerald.britton@gmail.com"],
    # load_on_reg = True
)

# ------------------------------------------------------------------------
#
# libmapservice
#
# ------------------------------------------------------------------------
register(
    GENERAL,
    id="libmapservice",
    name="mapservice lib",
    description=_("Provides base functionality for map services."),
    version="1.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="libmapservice.py",
    authors=["The Gramps project"],
    authors_email=["http://gramps-project.org"],
)

# ------------------------------------------------------------------------
#
# libnarrate
#
# ------------------------------------------------------------------------
register(
    GENERAL,
    id="libnarrate",
    name="narration lib",
    description=_("Provides Textual Narration."),
    version="1.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="libnarrate.py",
    authors=["Brian Matherly"],
    authors_email=["brian@gramps-project.org"],
)

# ------------------------------------------------------------------------
#
# libodfbackend
#
# ------------------------------------------------------------------------
register(
    GENERAL,
    id="libodfbackend",
    name="odfbackend lib",
    description=_("Manages an ODF file implementing DocBackend."),
    version="1.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="libodfbackend.py",
    authors=["The Gramps project"],
    authors_email=["http://gramps-project.org"],
)

# ------------------------------------------------------------------------
#
# libpersonview
#
# ------------------------------------------------------------------------
register(
    GENERAL,
    id="libpersonview",
    name="person list lib",
    description=_("Provides the Base needed for the List People views."),
    version="1.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="libpersonview.py",
    authors=["The Gramps project"],
    authors_email=["http://gramps-project.org"],
)

# ------------------------------------------------------------------------
#
# libprogen
#
# ------------------------------------------------------------------------
register(
    GENERAL,
    id="libprogen",
    name="Pro-Gen lib",
    description=_("Provides common functionality for Pro-Gen import"),
    version="1.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="libprogen.py",
    authors=["The Gramps project"],
    authors_email=["http://gramps-project.org"],
    # load_on_reg = True
)

# ------------------------------------------------------------------------
#
# libplaceview
#
# ------------------------------------------------------------------------
register(
    GENERAL,
    id="libplaceview",
    name="place list lib",
    description=_("Provides the Base needed for the List Place views."),
    version="1.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="libplaceview.py",
    authors=["The Gramps project"],
    authors_email=["http://gramps-project.org"],
)

# ------------------------------------------------------------------------
#
# libsubstkeyword
#
# ------------------------------------------------------------------------
register(
    GENERAL,
    id="libsubstkeyword",
    name="Substitution Values",
    description=_("Provides variable substitution on display lines."),
    version="1.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="libsubstkeyword.py",
    authors=["The Gramps project"],
    authors_email=["http://gramps-project.org"],
)

# ------------------------------------------------------------------------
#
# libtreebase
#
# ------------------------------------------------------------------------
register(
    GENERAL,
    id="libtreebase",
    name="Graphical report lib",
    description=_(
        "Provides the base needed for the ancestor and " "descendant graphical reports."
    ),
    version="1.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="libtreebase.py",
    authors=["The Gramps project"],
    authors_email=["http://gramps-project.org"],
)
